__all__ = (
    "get_current_footprint",
    "generate_all_sky",
    "band_count_ratios",
    "SkyAreaGenerator",
    "SkyAreaGeneratorGalplane",
    "EuclidOverlapFootprint",
    "CurrentAreaMap",
    "Phase3AreaMap",
)

import os
import warnings

import astropy.units as u
import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord
from numpy.lib import recfunctions as rfn
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon

from rubin_scheduler import data as rs_data
from rubin_scheduler.utils import DEFAULT_NSIDE, Site, _angular_separation, _hpid2_ra_dec, angular_separation

from .utils import IntRounded


def get_current_footprint(nside):
    """Convenience method to return the current footprint.

    This is primarily a way to help rubin-sim users keep up to date
    on footprint changes (as we have been moving to new subclasses).

    Parameters
    ----------
    nside : `int`
        The nside for the footprint map.

    Returns
    -------
    footprint_arrays, label_array : `np.array`, (N,), `np.array`, (N,)
        HEALPix target survey maps for ugrizy,
        and array of string labels for each healpix to indicate the "region".
    """
    sky = CurrentAreaMap(nside=nside)
    footprints, labels = sky.return_maps()
    return footprints, labels


def generate_all_sky(nside=DEFAULT_NSIDE, elevation_limit=20, mask=hp.UNSEEN):
    """Set up a healpix map over the entire sky.
    Calculate RA & Dec, Galactic l & b, Ecliptic l & b, for all healpixels.
    Calculate max altitude, to define areas which LSST cannot reach.

    This is intended to be a useful tool to use to set up target maps,
    beyond the standard maps provided in the various SkyArea generator maps.
    Masking based on RA, Dec, Galactic or Ecliptic lat and lon is easier.

    Parameters
    ----------
    nside : `int`, optional
        Resolution for the healpix maps.
        Default None uses rubin_scheduler.scheduler.utils.set_default_nside
        to set default (often 32).
    elevation_limit : `float`, optional
        Elevation limit for map.
        Parts of the sky which do not reach this elevation limit
        will be set to `mask.`
    mask : `float`, optional
        Mask value for 'unreachable' parts of the sky,
        defined as elevation < 20.

    Returns
    -------
    maps : `dict` {`str`: `np.ndarray`, (N,)}
        A dictionary of `map` (the skymap healpix array, with `mask` values),
        `ra`, `dec`, eclip_lat`, `eclip_lon`, `gal_lat`, `gal_lon` values.
        All coordinates are in radians.
    """

    # Calculate coordinates of everything.
    skymap = np.zeros(hp.nside2npix(nside), float)
    ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
    coord = SkyCoord(ra=ra * u.rad, dec=dec * u.rad, frame="icrs")
    eclip_lat = coord.barycentrictrueecliptic.lat.deg
    eclip_lon = coord.barycentrictrueecliptic.lon.deg
    gal_lon = coord.galactic.l.deg
    gal_lat = coord.galactic.b.deg

    # Calculate max altitude (when on meridian).
    lsst_site = Site("LSST")
    elev_max = np.pi / 2.0 - np.abs(dec - lsst_site.latitude_rad)
    skymap = np.where(IntRounded(elev_max) >= IntRounded(np.radians(elevation_limit)), skymap, mask)

    return {
        "map": skymap,
        "ra": np.degrees(ra),
        "dec": np.degrees(dec),
        "eclip_lat": eclip_lat,
        "eclip_lon": eclip_lon,
        "gal_lat": gal_lat,
        "gal_lon": gal_lon,
    }


def band_count_ratios(target_maps):
    """Compute the desired ratio of observations per band.

    Given a goal map that includes multiple bands, sum the number of
    pixels in each map and return this per-band, normalized so the sum
    across all bands is 1.
    If the map is constant over all healpixels, this is the
    ratio of bands at the max/constant value.
    """
    results = {}
    all_norm = 0.0
    for key in target_maps:
        good = target_maps[key] > 0
        results[key] = np.sum(target_maps[key][good])
        all_norm += results[key]
    for key in results:
        results[key] /= all_norm
    return results


class SkyAreaGenerator:
    """
    Generate survey footprint maps in each band.

    Parameters
    ----------
    nside : `int`
        Healpix nside
    dust_limit : `float`
        E(B-V) limit for dust extinction. Default of 0.199.
    smoothing_cutoff : `float`
       We apply a smoothing band to the defined dust-free region to
       avoid sharp edges. Larger values = less area, but guaranteed
       less dust extinction. Reflects the value to cut at, after smoothing.
    smoothing_beam : `float`
        The size of the smoothing band, in degrees.
    lmc_ra, lmc_dec : `float`, `float`
        RA and Dec locations of the LMC, in degrees.
    lmc_radius : `float`
        The radius to use around the LMC, in degrees.
    smc_ra, smc_dec : `float`, `float`
        RA and Dec locations for the center of the SMC, in degrees.
    smc_radius : `float`
        The radius to use around the SMC, degrees.
    scp_dec_max : `float`
        Maximum declination for the south celestial pole region, degrees.
    gal_long1 : `float`
        Longitude at which to start the GP region, in degrees.
    gal_long2 : `float`
        Longitude at which to stop the GP region, degrees.
        Order matters for gal_long1 / gal_long2!
    gal_lat_width_max : `float`
        Max width of the galactic plane, in degrees.
    center_width : `float`
        Width at the center of the galactic plane region, in degrees.
    end_width: `float`
        Width at the remainder of the galactic plane region, in degrees.
    gal_dec_max : `float`
        Maximum declination for the galactic plane region, degrees.
    dusty_dec_min : `float`
        The minimum dec for the dusty plane region, in degrees.
    dusty_dec_max : `float`
        The maximum dec for the dusty plane, degrees.
    eclat_min : `float`
        Ecliptic latitutde minimum for the NES, degrees.
    eclat_max : `float`
        Ecliptic latitude maximum for the NES, degrees.
    eclip_dec_min : `float`
        Declination minimum for the NES, degrees.
    nes_glon_limit : `float`
        Galactic longitude limit for the NES, degrees.
    virgo_ra, virgo_dec : `float`, `float`
        RA and Dec values for the Virgo coverage center, in degrees.
    virgo_radius : `float`
        Radius for the virgo coverage, in degrees.
    """

    def __init__(
        self,
        nside=DEFAULT_NSIDE,
        dust_limit=0.199,
        smoothing_cutoff=0.45,
        smoothing_beam=10,
        lmc_ra=80.893860,
        lmc_dec=-69.756126,
        lmc_radius=8,
        smc_ra=13.186588,
        smc_dec=-72.828599,
        smc_radius=5,
        scp_dec_max=-60,
        gal_long1=335,
        gal_long2=25,
        gal_lat_width_max=23,
        center_width=12,
        end_width=4,
        gal_dec_max=12,
        low_dust_dec_min=-70,
        low_dust_dec_max=15,
        adjust_halves=12,
        dusty_dec_min=-90,
        dusty_dec_max=15,
        eclat_min=-10,
        eclat_max=10,
        eclip_dec_min=0,
        nes_glon_limit=45.0,
        virgo_ra=186.75,
        virgo_dec=12.717,
        virgo_radius=8.75,
    ):
        self.nside = nside
        self.hpid = np.arange(0, hp.nside2npix(nside))
        self.read_dustmap()

        self.lmc_ra = lmc_ra
        self.lmc_dec = lmc_dec
        self.lmc_radius = lmc_radius
        self.smc_ra = smc_ra
        self.smc_dec = smc_dec
        self.smc_radius = smc_radius

        self.virgo_ra = virgo_ra
        self.virgo_dec = virgo_dec
        self.virgo_radius = virgo_radius

        self.scp_dec_max = scp_dec_max

        self.gal_long1 = gal_long1
        self.gal_long2 = gal_long2
        self.gal_lat_width_max = gal_lat_width_max
        self.center_width = center_width
        self.end_width = end_width
        self.gal_dec_max = gal_dec_max

        self.low_dust_dec_min = low_dust_dec_min
        self.low_dust_dec_max = low_dust_dec_max
        self.adjust_halves = adjust_halves

        self.dusty_dec_min = dusty_dec_min
        self.dusty_dec_max = dusty_dec_max

        self.eclat_min = eclat_min
        self.eclat_max = eclat_max
        self.eclip_dec_min = eclip_dec_min
        self.nes_glon_limit = nes_glon_limit

        # Ra/dec in degrees and other coordinates
        self.ra, self.dec = hp.pix2ang(nside, self.hpid, lonlat=True)
        self.coord = SkyCoord(ra=self.ra * u.deg, dec=self.dec * u.deg, frame="icrs")
        self.eclip_lat = self.coord.barycentrictrueecliptic.lat.deg
        self.eclip_lon = self.coord.barycentrictrueecliptic.lon.deg
        self.gal_lon = self.coord.galactic.l.deg
        self.gal_lat = self.coord.galactic.b.deg

        # Set the low extinction area
        self.low_dust = np.where((self.dustmap < dust_limit), 1, 0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            self.low_dust = hp.smoothing(self.low_dust, fwhm=np.radians(smoothing_beam))
        self.low_dust = np.where(self.low_dust > smoothing_cutoff, 1, 0)

    def read_dustmap(self, dustmap_file=None):
        """Read the dustmap from rubin_scheduler, in the
        appropriate resolution."""
        # Dustmap from rubin_sim_data
        if dustmap_file is None:
            datadir = rs_data.get_data_dir()
            if datadir is None:
                raise Exception('Cannot find datadir, please set "RUBIN_SIM_DATA_DIR"')
            datadir = os.path.join(datadir, "scheduler", "dust_maps")
            filename = os.path.join(datadir, "dust_nside_%i.npz" % self.nside)
        with np.load(filename) as data:
            self.dustmap = data["ebvMap"]

    def _set_circular_region(self, ra_center, dec_center, radius):
        """Define a circular region centered on ra_center, dec_center."""
        # find the healpixels that cover a circle of radius
        # "radius" around ra/dec center (deg).
        result = np.zeros(len(self.ra))
        distance = _angular_separation(
            np.radians(ra_center),
            np.radians(dec_center),
            np.radians(self.ra),
            np.radians(self.dec),
        )
        result[np.where(distance < np.radians(radius))] = 1
        return result

    def _set_bulge_diamond(self, center_width, end_width, gal_long1, gal_long2):
        """
        Define a Galactic Bulge diamond-ish region.

        Parameters
        ----------
        center_width : `float`
            Width at the center of the galactic plane region.
        end_width : `float`
            Width at the remainder of the galactic plane region.
        gal_long1 : `float`
            Longitude at which to start the GP region.
        gal_long2 : `float`
            Longitude at which to stop the GP region.
            Order matters for gal_long1 / gal_long2!

        Returns
        -------
        bulge : `np.ndarray`
            HEALpix array with 1 for bulge pixels, 0 otherwise
        """
        # Reject anything beyond the central width.
        bulge = np.where(np.abs(self.gal_lat) < center_width, 1, 0)
        # Apply the galactic longitude cuts,
        # so that plane goes between gal_long1 to gal_long2.
        # This is NOT the shortest distance between the angles.
        gp_length = (gal_long2 - gal_long1) % 360
        # If the length is greater than 0 then we can add additional cuts.
        if gp_length > 0:
            # First, remove anything outside the gal_long1/gal_long2 region.
            bulge = np.where(((self.gal_lon - gal_long1) % 360) < gp_length, bulge, 0)
            # Add the tapers.
            # These slope from the center (gp_center @ center_width)
            # to the edges (gp_center + gp_length/2 @ end_width).
            half_width = gp_length / 2.0
            slope = (center_width - end_width) / half_width
            # The 'center' can have a wrap-around 0 problem
            gp_center = (gal_long1 + half_width) % 360
            # Calculate the longitude-distance btwn any point and the 'center'
            gp_dist = (self.gal_lon - gp_center) % 360
            gp_dist = np.abs(np.where((gp_dist > 180), (180 - gp_dist) % 180, gp_dist))
            lat_limit = np.abs(center_width - slope * gp_dist)
            bulge = np.where((np.abs(self.gal_lat)) < lat_limit, bulge, 0)
        return bulge

    def _set_bulge_rectangle(self, lat_width, gal_long1, gal_long2):
        """
        Define a Galactic Bulge as a simple rectangle in galactic coordinates,
        centered on the plane

        Parameters
        ----------
        lat_width : `float`
            Latitude with for the galactic bulge
        gal_long1 : float
            Longitude at which to start the GP region.
        gal_long2 : float
            Longitude at which to stop the GP region.
            Order matters for gal_long1 / gal_long2!

        Returns
        -------
        bulge : `np.ndarray`
            HEALpix array with 1 for bulge pixels, 0 otherwise
        """
        bulge = np.where(np.abs(self.gal_lat) < lat_width, 1, 0)
        # This is NOT the shortest distance between the angles.
        gp_length = (gal_long2 - gal_long1) % 360
        # If the length is greater than 0 then we can add additional cuts.
        if gp_length > 0:
            # First, remove anything outside the gal_long1/gal_long2 region.
            bulge = np.where(((self.gal_lon - gal_long1) % 360) < gp_length, bulge, 0)
        return bulge

    def add_virgo_cluster(self, band_ratios, label="virgo"):
        """Define a circular region around the Virgo Cluster.
        Updates self.healmaps and self.pix_labels

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        temp_map = np.zeros(hp.nside2npix(self.nside))
        temp_map += self._set_circular_region(self.virgo_ra, self.virgo_dec, self.virgo_radius)
        # Don't overide any pixels that have already been designated
        indx = np.where((temp_map > 0) & (self.pix_labels == ""))
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_magellanic_clouds(
        self,
        band_ratios,
        label="LMC_SMC",
    ):
        """Define circular regions around the Magellanic Clouds.
        Updates self.healmaps and self.pix_labels.

        Parameters
        -----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        temp_map = np.zeros(hp.nside2npix(self.nside))
        # Define the LMC pixels
        temp_map += self._set_circular_region(self.lmc_ra, self.lmc_dec, self.lmc_radius)
        # Define the SMC pixels
        temp_map += self._set_circular_region(self.smc_ra, self.smc_dec, self.smc_radius)
        # Add a simple bridge between the two - to remove the gap
        mc_dec_min = self.dec[np.where(temp_map > 0)].min()
        mc_dec_max = self.dec[np.where(temp_map > 0)].max()
        temp_map += np.where(
            ((self.ra > self.smc_ra) & (self.ra < self.lmc_ra))
            & ((self.dec > mc_dec_min) & (self.dec < mc_dec_max)),
            1,
            0,
        )

        # Don't overide any pixels that have already been designated
        indx = np.where((temp_map > 0) & (self.pix_labels == ""))

        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_scp(self, band_ratios, label="scp"):
        """Define a south celestial pole cap region.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        indx = np.where((self.dec < self.scp_dec_max) & (self.pix_labels == ""))

        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_bulge(self, band_ratios, label="bulge"):
        """Define a bulge region, where the 'bulge' is a large "diamond"
        centered on the galactic center.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        b1 = self._set_bulge_diamond(
            center_width=self.center_width,
            end_width=self.end_width,
            gal_long1=self.gal_long1,
            gal_long2=self.gal_long2,
        )
        b2 = self._set_bulge_rectangle(self.gal_lat_width_max, self.gal_long1, self.gal_long2)
        b2[np.where(self.gal_lat > 0)] = 0

        bulge = b1 + b2

        bulge[np.where(np.abs(self.gal_lat) > self.gal_lat_width_max)] = 0
        indx = np.where((bulge > 0) & (self.pix_labels == ""))
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_lowdust_wfd(self, band_ratios, label="lowdust"):
        """Define a low-dust WFD region.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        dustfree = np.where(
            (self.dec > self.low_dust_dec_min) & (self.dec < self.low_dust_dec_max) & (self.low_dust == 1),
            1,
            0,
        )

        dustfree[np.where(self.low_dust == 0)] = 0

        if self.adjust_halves > 0:
            dustfree = np.where(
                (self.gal_lat < 0) & (self.dec > self.low_dust_dec_max - self.adjust_halves),
                0,
                dustfree,
            )

        indx = np.where((dustfree > 0) & (self.pix_labels == ""))
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_dusty_plane(self, band_ratios, label="dusty_plane"):
        """Define high-dust region of the map.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        dusty = np.where(
            ((self.dec > self.dusty_dec_min) & (self.dec < self.dusty_dec_max) & (self.low_dust == 0)),
            1,
            0,
        )

        indx = np.where((dusty > 0) & (self.pix_labels == ""))
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def add_nes(self, band_ratios, label="nes"):
        """Define a North Ecliptic Plane region.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        nes = np.where(
            ((self.eclip_lat > self.eclat_min) | (self.dec > self.eclip_dec_min))
            & (self.eclip_lat < self.eclat_max),
            1,
            0,
        )

        nes[np.where(self.gal_lon < self.nes_glon_limit)] = 0
        nes[np.where(self.gal_lon > (360 - self.nes_glon_limit))] = 0

        indx = np.where((nes > 0) & (self.pix_labels == ""))
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def return_maps(
        self,
        magellenic_clouds_ratios={
            "u": 0.32,
            "g": 0.4,
            "r": 1.0,
            "i": 1.0,
            "z": 0.9,
            "y": 0.9,
        },
        scp_ratios={"u": 0.1, "g": 0.1, "r": 0.1, "i": 0.1, "z": 0.1, "y": 0.1},
        nes_ratios={"g": 0.28, "r": 0.4, "i": 0.4, "z": 0.28},
        dusty_plane_ratios={
            "u": 0.1,
            "g": 0.28,
            "r": 0.28,
            "i": 0.28,
            "z": 0.28,
            "y": 0.1,
        },
        low_dust_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        bulge_ratios={"u": 0.18, "g": 1.0, "r": 1.05, "i": 1.05, "z": 1.0, "y": 0.23},
        virgo_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
    ):
        """
        Return the survey sky maps and labels.

        Parameters
        ----------
        magellanic_clouds_ratios :  `dict` {`str`: `float`}
            Magellanic clouds band ratios.
        scp_ratios :  `dict` {`str`: `float`}
            SCP band ratios.
        nes_ratios :  `dict` {`str`: `float`}
            NES band ratios
        dusty_plane-ratios :  `dict` {`str`: `float`}
            dusty plane band ratios
        low_dust_ratios :  `dict` {`str`: `float`}
            Low Dust WFD band ratios.
        bulge_ratios :  `dict` {`str`: `float`}
            Bulge region band ratios.
        virgo_ratios :  `dict` {`str`: `float`}
            Virgo cluster coverage band ratios.

        Returns
        --------
        self.healmaps, self.pix_labels : `np.ndarray`, (N,), `np.ndarray`, (N,)
            HEALPix target survey maps for ugrizy,
            and string labels for each healpix to indicate the "region".


        Notes
        -----
        Each healpix point can only belong to one region. Which region it is
        assigned to first will be used for its definition, thus order
        matters within this method. The region defines the band ratios.

        The band ratios contain information about the *ratio* of visits
        in that region (compared to some reference point in the entire map)
        in each particular band. By convention, the low_dust_wfd ratio
        in r band is set to "1" and all other values are then in reference
        to that.

        For example: if scp_ratios['u'] = 0.1 and the low_dust_wfd['r'] = 1,
        then when the low-dust WFD has 10 visits in r band, the SCP should
        have obtained 1 visits in u band (per pixel).
        """

        # Array to hold the labels for each pixel
        self.pix_labels = np.zeros(hp.nside2npix(self.nside), dtype="U20")
        self.healmaps = np.zeros(
            hp.nside2npix(self.nside),
            dtype=list(zip(["u", "g", "r", "i", "z", "y"], [float] * 7)),
        )

        # Note, order here matters.
        # Once a HEALpix is set and labled, subsequent add_ methods
        # will not override that pixel.
        self.add_magellanic_clouds(magellenic_clouds_ratios)
        self.add_lowdust_wfd(low_dust_ratios)
        self.add_virgo_cluster(virgo_ratios)
        self.add_bulge(bulge_ratios)
        self.add_nes(nes_ratios)
        self.add_dusty_plane(dusty_plane_ratios)
        self.add_scp(scp_ratios)

        return self.healmaps, self.pix_labels

    def estimate_visits(self, nvis_total, fov_area=9.6, **kwargs):
        """Convience method for converting relative maps into number of visits

        Parameters
        ----------
        nvis_total : `int`
            The total number of visits in the survey
        fov_area : `float`
            The area of a single visit (sq degrees)
        **kwargs :
            Gets passed to self.return_maps if one wants to change the
            default ratios.

        Returns
        -------
        result : `np.array`, (N,)
            array with bandname dtypes that have HEALpix arrays with the
            number of expected visits of each HEALpix center
        sum_map : `np.array`, (N,)
            The number of visits summed over all the bands
        labels : `np.ndarray`, (N,)
            Array string labels for each HEALpix
        """
        # Note that there really ought to be a fudge factor here;
        # typically the number of visits per pixel will be slightly
        # more than requested in the map, due to dithering.
        # The fudge factor will depend on the complexity of the region's
        # boundaries, but can be on the order of 1.3.
        healmaps, labels = self.return_maps(**kwargs)

        sum_map = rfn.structured_to_unstructured(healmaps).sum(axis=1)

        norm = np.sum(sum_map)
        pix_area = hp.nside2pixarea(self.nside, degrees=True)
        pix_per_visit = fov_area / pix_area

        result = np.zeros_like(healmaps)
        for key in result.dtype.names:
            result[key] = healmaps[key] / norm * pix_per_visit * nvis_total

        return result, sum_map / norm * pix_per_visit * nvis_total, labels

    def estimate_visits_per_label(self, nvis_total, **kwargs):
        """Estimate how many visits would be used for each region

        Parameters
        ----------
        nvis_total : `int`
            The total number of visits in the survey
        **kwargs :
            Gets passed to self.return_maps if one wants to change the
            default ratios.

        Returns
        -------
        result : `dict`
            Dictionary with keys that are label names and values that are the
            expected number of visits for that region if nvis_total is reached.
        """

        healmaps, labels = self.return_maps(**kwargs)
        sum_map = rfn.structured_to_unstructured(healmaps).sum(axis=1)
        ulabels = np.unique(labels)
        label_sums = {}
        norm = 0
        for label in ulabels:
            in_region = np.where(labels == label)
            label_sums[label] = sum_map[in_region].sum()
            norm += label_sums[label]
        result = {}
        for key in ulabels:
            result[key] = label_sums[key] / norm * nvis_total
        return result


class SkyAreaGeneratorGalplane(SkyAreaGenerator):
    """
    Generate survey footprint maps in each band.
    Adds a 'bulgy' galactic plane coverage map.

    Parameters
    ----------
    nside : `int`
        Healpix nside
    dust_limit : `float`
        E(B-V) limit for dust extinction. Default of 0.199.
    smoothing_cutoff : `float`
       We apply a smoothing band to the defined dust-free region to
       avoid sharp edges. Larger values = less area, but guaranteed
       less dust extinction. Reflects the value to cut at, after smoothing.
    smoothing_beam : `float`
        The size of the smoothing band, in degrees.
    lmc_ra, lmc_dec : `float`, `float`
        RA and Dec locations of the LMC, in degrees.
    lmc_radius : `float`
        The radius to use around the LMC, in degrees.
    smc_ra, smc_dec : `float`, `float`
        RA and Dec locations for the center of the SMC, in degrees.
    smc_radius : `float`
        The radius to use around the SMC, degrees.
    scp_dec_max : `float`
        Maximum declination for the south celestial pole region, degrees.
    gal_long1 : `float`
        Longitude at which to start the GP region, in degrees.
    gal_long2 : `float`
        Longitude at which to stop the GP region, degrees.
        Order matters for gal_long1 / gal_long2!
    gal_lat_width_max : `float`
        Max width of the galactic plane, in degrees.
    center_width : `float`
        Width at the center of the galactic plane region, in degrees.
    end_width: `float`
        Width at the remainder of the galactic plane region, in degrees.
    gal_dec_max : `float`
        Maximum declination for the galactic plane region, degrees.
    dusty_dec_min : `float`
        The minimum dec for the dusty plane region, in degrees.
    dusty_dec_max : `float`
        The maximum dec for the dusty plane, degrees.
    eclat_min : `float`
        Ecliptic latitutde minimum for the NES, degrees.
    eclat_max : `float`
        Ecliptic latitude maximum for the NES, degrees.
    eclip_dec_min : `float`
        Declination minimum for the NES, degrees.
    nes_glon_limit : `float`
        Galactic longitude limit for the NES, degrees.
    virgo_ra, virgo_dec : `float`, `float`
        RA and Dec values for the Virgo coverage center, in degrees.
    virgo_radius : `float`
        Radius for the virgo coverage, in degrees.
    """

    def __init__(self, lmc_ra=89.0, lmc_dec=-70, **kwargs):
        super().__init__(lmc_ra=lmc_ra, lmc_dec=lmc_dec, **kwargs)

    def add_bulgy(self, band_ratios, label="bulgy"):
        """Define a bulge region, where the 'bulge' is a series of
        circles set by points defined to match as best as possible the
        map requested by the SMWLV working group on galactic plane coverage.
        Implemented in v3.0.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        # Some RA, dec, radius points that
        # seem to cover the areas that are desired
        points = [
            [100.90, 9.55, 3],
            [84.92, -5.71, 3],
            [288.84, 9.18, 3.8],
            [266.3, -29, 14.5],
            [279, -13, 10],
            [256, -45, 5],
            [155, -56.5, 6.5],
            [172, -62, 5],
            [190, -65, 5],
            [210, -64, 5],
            [242, -58, 5],
            [225, -60, 6.5],
        ]
        for point in points:
            dist = angular_separation(self.ra, self.dec, point[0], point[1])
            # Only change pixels where the label isn't already set.
            indx = np.where((dist < point[2]) & (self.pix_labels == ""))
            self.pix_labels[indx] = label
            for bandname in band_ratios:
                self.healmaps[bandname][indx] = band_ratios[bandname]

    def return_maps(
        self,
        magellenic_clouds_ratios={
            "u": 0.32,
            "g": 0.4,
            "r": 1.0,
            "i": 1.0,
            "z": 0.9,
            "y": 0.9,
        },
        low_dust_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        virgo_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        scp_ratios={"u": 0.08, "g": 0.15, "r": 0.08, "i": 0.15, "z": 0.08, "y": 0.06},
        nes_ratios={"g": 0.23, "r": 0.33, "i": 0.33, "z": 0.23},
        bulge_ratios={"u": 0.19, "g": 0.57, "r": 1.15, "i": 1.05, "z": 0.78, "y": 0.57},
        dusty_plane_ratios={
            "u": 0.07,
            "g": 0.13,
            "r": 0.28,
            "i": 0.28,
            "z": 0.25,
            "y": 0.18,
        },
    ):
        """
        Return the survey sky maps and labels.

        Parameters
        ----------
        magellanic_clouds_ratios :  `dict` {`str`: `float`}
            Magellanic clouds band ratios.
        scp_ratios :  `dict` {`str`: `float`}
            SCP band ratios.
        nes_ratios :  `dict` {`str`: `float`}
            NES band ratios
        dusty_plane-ratios :  `dict` {`str`: `float`}
            dusty plane band ratios
        low_dust_ratios :  `dict` {`str`: `float`}
            Low Dust WFD band ratios.
        bulge_ratios :  `dict` {`str`: `float`}
            Bulge region band ratios (note this is the 'bulgy' bulge).
        virgo_ratios :  `dict` {`str`: `float`}
            Virgo cluster coverage band ratios.

        Returns
        --------
        self.healmaps, self.pix_labels : `np.ndarray`, (N,), `np.ndarray`, (N,)
            HEALPix target survey maps for ugrizy,
            and string labels for each healpix to indicate the "region".


        Notes
        -----
        Each healpix point can only belong to one region. Which region it is
        assigned to first will be used for its definition, thus order
        matters within this method. The region defines the band ratios.

        The band ratios contain information about the *ratio* of visits
        in that region (compared to some reference point in the entire map)
        in each particular band. By convention, the low_dust_wfd ratio
        in r band is set to "1" and all other values are then in reference
        to that.

        For example: if scp_ratios['u'] = 0.1 and the low_dust_wfd['r'] = 1,
        then when the low-dust WFD has 10 visits in r band, the SCP should
        have obtained 1 visits in u band (per pixel).
        """
        self.pix_labels = np.zeros(hp.nside2npix(self.nside), dtype="U20")
        dt = list(zip(["u", "g", "r", "i", "z", "y"], [float] * 7))
        self.healmaps = np.zeros(hp.nside2npix(self.nside), dtype=dt)

        self.add_magellanic_clouds(magellenic_clouds_ratios)
        self.add_lowdust_wfd(low_dust_ratios)
        self.add_virgo_cluster(virgo_ratios)
        self.add_bulgy(bulge_ratios)
        self.add_nes(nes_ratios)
        self.add_dusty_plane(dusty_plane_ratios)
        self.add_scp(scp_ratios)

        return self.healmaps, self.pix_labels


class EuclidOverlapFootprint(SkyAreaGeneratorGalplane):
    """
    Generate survey footprint maps in each band.
    This uses a bulgy coverage in the galactic plane, plus small
    Euclid footprint extension to the low-dust WFD.

    Parameters
    ----------
    nside : `int`
        Healpix nside
    dust_limit : `float`
        E(B-V) limit for dust extinction. Default of 0.199.
    smoothing_cutoff : `float`
       We apply a smoothing band to the defined dust-free region to
       avoid sharp edges. Larger values = less area, but guaranteed
       less dust extinction. Reflects the value to cut at, after smoothing.
    smoothing_beam : `float`
        The size of the smoothing band, in degrees.
    lmc_ra, lmc_dec : `float`, `float`
        RA and Dec locations of the LMC, in degrees.
    lmc_radius : `float`
        The radius to use around the LMC, in degrees.
    smc_ra, smc_dec : `float`, `float`
        RA and Dec locations for the center of the SMC, in degrees.
    smc_radius : `float`
        The radius to use around the SMC, degrees.
    scp_dec_max : `float`
        Maximum declination for the south celestial pole region, degrees.
    gal_long1 : `float`
        Longitude at which to start the GP region, in degrees.
    gal_long2 : `float`
        Longitude at which to stop the GP region, degrees.
        Order matters for gal_long1 / gal_long2!
    gal_lat_width_max : `float`
        Max width of the galactic plane, in degrees.
    center_width : `float`
        Width at the center of the galactic plane region, in degrees.
    end_width: `float`
        Width at the remainder of the galactic plane region, in degrees.
    gal_dec_max : `float`
        Maximum declination for the galactic plane region, degrees.
    dusty_dec_min : `float`
        The minimum dec for the dusty plane region, in degrees.
    dusty_dec_max : `float`
        The maximum dec for the dusty plane, degrees.
    eclat_min : `float`
        Ecliptic latitutde minimum for the NES, degrees.
    eclat_max : `float`
        Ecliptic latitude maximum for the NES, degrees.
    eclip_dec_min : `float`
        Declination minimum for the NES, degrees.
    nes_glon_limit : `float`
        Galactic longitude limit for the NES, degrees.
    virgo_ra, virgo_dec : `float`, `float`
        RA and Dec values for the Virgo coverage center, in degrees.
    virgo_radius : `float`
        Radius for the virgo coverage, in degrees.
    euclid_contour_file : `str`
        File containing a Euclid footprint contour file.
        Default of none uses the file in rubin_sim_data/scheduler.
    """

    def __init__(self, euclid_contour_file=None, lmc_radius=6, smc_radius=4, **kwargs):
        super().__init__(lmc_radius=lmc_radius, smc_radius=smc_radius, **kwargs)
        self.euclid_contour_file = euclid_contour_file

    def add_euclid_overlap(
        self,
        band_ratios,
        label="euclid_overlap",
    ):
        """Define a small extension (few degrees) to the low-dust WFD
        to accomodate the Euclid footprint.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        names = ["RA", "dec"]
        types = [float, float]
        if self.euclid_contour_file is None:
            self.euclid_contour_file = os.path.join(
                rs_data.get_data_dir(), "scheduler/EWS.SGC.Mainland.ROI.2022.RADEC.txt"
            )
        euclid_contours = np.genfromtxt(self.euclid_contour_file, dtype=list(zip(names, types)))

        wrap_ra = self.ra + 0
        wrap_ra[np.where(wrap_ra > 180)] -= 360

        polygon = Polygon(zip(euclid_contours["RA"], euclid_contours["dec"]))
        in_poly = [polygon.contains(Point(x, y)) for x, y in zip(wrap_ra, self.dec)]

        # find which map points are inside the contour
        indx = np.where((np.array(in_poly)) & (self.pix_labels == ""))[0]
        self.pix_labels[indx] = label
        for bandname in band_ratios:
            self.healmaps[bandname][indx] = band_ratios[bandname]

    def return_maps(
        self,
        magellenic_clouds_ratios={
            "u": 0.32,
            "g": 0.4,
            "r": 1.0,
            "i": 1.0,
            "z": 0.9,
            "y": 0.9,
        },
        low_dust_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        virgo_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        scp_ratios={"u": 0.08, "g": 0.15, "r": 0.08, "i": 0.15, "z": 0.08, "y": 0.06},
        nes_ratios={"g": 0.23, "r": 0.33, "i": 0.33, "z": 0.23},
        bulge_ratios={"u": 0.19, "g": 0.57, "r": 1.15, "i": 1.05, "z": 0.78, "y": 0.57},
        dusty_plane_ratios={
            "u": 0.07,
            "g": 0.13,
            "r": 0.28,
            "i": 0.28,
            "z": 0.25,
            "y": 0.18,
        },
        euclid_ratios={"u": 0.32, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
    ):
        """
        Return the survey sky maps and labels.

        Parameters
        ----------
        magellanic_clouds_ratios :  `dict` {`str`: `float`}
            Magellanic clouds band ratios.
        scp_ratios :  `dict` {`str`: `float`}
            SCP band ratios.
        nes_ratios :  `dict` {`str`: `float`}
            NES band ratios
        dusty_plane-ratios :  `dict` {`str`: `float`}
            dusty plane band ratios
        low_dust_ratios :  `dict` {`str`: `float`}
            Low Dust WFD band ratios.
        bulge_ratios :  `dict` {`str`: `float`}
            Bulge region band ratios (note this is the 'bulgy' bulge).
        virgo_ratios :  `dict` {`str`: `float`}
            Virgo cluster coverage band ratios.
        euclid_ratios : `dict` {`str`: `float`}
            Euclid footprint overlap ratios.

        Returns
        --------
        self.healmaps, self.pix_labels : `np.ndarray`, (N,), `np.ndarray`, (N,)
            HEALPix target survey maps for ugrizy,
            and string labels for each healpix to indicate the "region".


        Notes
        -----
        Each healpix point can only belong to one region. Which region it is
        assigned to first will be used for its definition, thus order
        matters within this method. The region defines the band ratios.

        The band ratios contain information about the *ratio* of visits
        in that region (compared to some reference point in the entire map)
        in each particular band. By convention, the low_dust_wfd ratio
        in r band is set to "1" and all other values are then in reference
        to that.

        For example: if scp_ratios['u'] = 0.1 and the low_dust_wfd['r'] = 1,
        then when the low-dust WFD has 10 visits in r band, the SCP should
        have obtained 1 visits in u band (per pixel).
        """

        # Array to hold the labels for each pixel
        self.pix_labels = np.zeros(hp.nside2npix(self.nside), dtype="U20")
        self.healmaps = np.zeros(
            hp.nside2npix(self.nside),
            dtype=list(zip(["u", "g", "r", "i", "z", "y"], [float] * 7)),
        )

        # Note, order here matters.
        # Once a HEALpix is set and labled, subsequent add_ methods
        # will not override that pixel.
        self.add_magellanic_clouds(magellenic_clouds_ratios)
        self.add_lowdust_wfd(low_dust_ratios)
        self.add_virgo_cluster(virgo_ratios)
        self.add_bulgy(bulge_ratios)
        self.add_nes(nes_ratios)
        self.add_dusty_plane(dusty_plane_ratios)
        self.add_euclid_overlap(euclid_ratios)
        self.add_scp(scp_ratios)

        return self.healmaps, self.pix_labels


class Phase3AreaMap(EuclidOverlapFootprint):
    def __init__(
        self,
        nside=DEFAULT_NSIDE,
        dust_limit=0.199,
        smoothing_cutoff=0.45,
        smoothing_beam=10,
        lmc_ra=80.893860,
        lmc_dec=-69.756126,
        lmc_radius=6,
        smc_ra=13.186588,
        smc_dec=-72.828599,
        smc_radius=4,
        scp_dec_max=-60,
        gal_long1=335,
        gal_long2=25,
        gal_lat_width_max=23,
        center_width=12,
        end_width=4,
        gal_dec_max=12,
        low_dust_dec_min=-70,
        low_dust_dec_max=15,
        adjust_halves=12,
        dusty_dec_min=-90,
        dusty_dec_max=15,
        eclat_min=-10,
        eclat_max=10,
        eclip_dec_min=0,
        nes_glon_limit=45.0,
        virgo_ra=186.75,
        virgo_dec=12.717,
        virgo_radius=8.75,
        euclid_contour_file=None,
    ):
        self.nside = nside
        self.hpid = np.arange(0, hp.nside2npix(nside))
        self.read_dustmap()

        self.lmc_ra = lmc_ra
        self.lmc_dec = lmc_dec
        self.lmc_radius = lmc_radius
        self.smc_ra = smc_ra
        self.smc_dec = smc_dec
        self.smc_radius = smc_radius

        self.virgo_ra = virgo_ra
        self.virgo_dec = virgo_dec
        self.virgo_radius = virgo_radius

        self.scp_dec_max = scp_dec_max

        self.gal_long1 = gal_long1
        self.gal_long2 = gal_long2
        self.gal_lat_width_max = gal_lat_width_max
        self.center_width = center_width
        self.end_width = end_width
        self.gal_dec_max = gal_dec_max

        self.low_dust_dec_min = low_dust_dec_min
        self.low_dust_dec_max = low_dust_dec_max
        self.adjust_halves = adjust_halves

        self.dusty_dec_min = dusty_dec_min
        self.dusty_dec_max = dusty_dec_max

        self.eclat_min = eclat_min
        self.eclat_max = eclat_max
        self.eclip_dec_min = eclip_dec_min
        self.nes_glon_limit = nes_glon_limit

        # Ra/dec in degrees and other coordinates
        self.ra, self.dec = hp.pix2ang(nside, self.hpid, lonlat=True)
        self.coord = SkyCoord(ra=self.ra * u.deg, dec=self.dec * u.deg, frame="icrs")
        self.eclip_lat = self.coord.barycentrictrueecliptic.lat.deg
        self.eclip_lon = self.coord.barycentrictrueecliptic.lon.deg
        self.gal_lon = self.coord.galactic.l.deg
        self.gal_lat = self.coord.galactic.b.deg

        # Set the low extinction area
        self.low_dust = np.where((self.dustmap < dust_limit), 1, 0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            self.low_dust = hp.smoothing(self.low_dust, fwhm=np.radians(smoothing_beam))
        self.low_dust = np.where(self.low_dust > smoothing_cutoff, 1, 0)

        self.euclid_contour_file = euclid_contour_file

    def add_bulgy(self, band_ratios, label="bulgy"):
        """Define a bulge region, where the 'bulge' is a series of
        circles set by points defined to match as best as possible the
        map requested by the SMWLV working group on galactic plane coverage.
        Implemented in v3.0.
        Updates self.healmaps and self.pix_labels.

        Parameters
        ----------
        band_ratios : `dict` {`str`: `float`}
            Dictionary of weights per band for the footprint.
        label : `str`, optional
            Label to apply to the resulting footprint
        """
        # Some RA, dec, radius points that
        # seem to cover the areas that are desired
        points = [
            [100.90, 9.55, 3],
            [84.92, -5.71, 3],
            [266.3, -29, 17],
            [279, -13, 10],
            [256, -45, 11],
            [155, -56.5, 6.5],
            [172, -62, 5],
            [190, -65, 5],
            [210, -64, 5],
            [242, -58, 6.5],
            [225, -60, 6.5],
        ]
        for point in points:
            dist = angular_separation(self.ra, self.dec, point[0], point[1])
            # Only change pixels where the label isn't already set.
            indx = np.where((dist < point[2]) & (self.pix_labels == ""))
            self.pix_labels[indx] = label
            for bandname in band_ratios:
                self.healmaps[bandname][indx] = band_ratios[bandname]

    def return_maps(
        self,
        magellenic_clouds_ratios={
            "u": 0.65,
            "g": 0.65,
            "r": 1.1,
            "i": 1.1,
            "z": 0.34,
            "y": 0.35,
        },
        scp_ratios={"u": 0.1, "g": 0.175, "r": 0.1, "i": 0.135, "z": 0.046, "y": 0.047},
        nes_ratios={"g": 0.255, "r": 0.33, "i": 0.33, "z": 0.23},
        dusty_plane_ratios={
            "u": 0.093,
            "g": 0.26,
            "r": 0.26,
            "i": 0.26,
            "z": 0.26,
            "y": 0.093,
        },
        low_dust_ratios={"u": 0.35, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        bulge_ratios={"u": 0.17, "g": 0.93, "r": 0.98, "i": 0.98, "z": 0.93, "y": 0.21},
        virgo_ratios={"u": 0.35, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
        euclid_ratios={"u": 0.35, "g": 0.4, "r": 1.0, "i": 1.0, "z": 0.9, "y": 0.9},
    ):
        # Array to hold the labels for each pixel
        self.pix_labels = np.zeros(hp.nside2npix(self.nside), dtype="U20")
        self.healmaps = np.zeros(
            hp.nside2npix(self.nside),
            dtype=list(zip(["u", "g", "r", "i", "z", "y"], [float] * 7)),
        )

        # Note, order here matters.
        # Once a HEALpix is set and labled, subsequent add_ methods
        # will not override that pixel.
        self.add_magellanic_clouds(magellenic_clouds_ratios)
        self.add_lowdust_wfd(low_dust_ratios)
        self.add_virgo_cluster(virgo_ratios)
        self.add_bulgy(bulge_ratios)
        self.add_nes(nes_ratios)
        self.add_dusty_plane(dusty_plane_ratios)
        self.add_euclid_overlap(euclid_ratios)
        self.add_scp(scp_ratios)

        return self.healmaps, self.pix_labels


class CurrentAreaMap(Phase3AreaMap):
    """Useful pointer so whatever the current standard footprint is."""

    pass



import unittest

import numpy as np
from astropy.time import Time

from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.schedulers import ComCamBandSched, DateSwapBandScheduler, SimpleBandSched
from rubin_scheduler.utils import SURVEY_START_MJD


class TestBandSchedulers(unittest.TestCase):
    def test_ComCamBandSched(self):
        illum_bins = np.arange(0, 100 + 1, 50)
        band_groups = (("g", "r", "i"), ("i", "z", "y"))
        bandsched = ComCamBandSched(illum_bins=illum_bins, loaded_band_groups=band_groups)
        mjd = SURVEY_START_MJD
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["g", "r", "i"])
        conditions.moon_phase_sunset = 40
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["g", "r", "i"])
        conditions.moon_phase_sunset = 60
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["i", "z", "y"])
        conditions.moon_phase_sunset = 100
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["i", "z", "y"])

    def test_comcambandsched_except(self):
        illum_bins = np.arange(0, 100 + 1, 25)
        band_groups = (("g", "r", "i"), ("i", "z", "y"))
        with self.assertRaises(ValueError):
            ComCamBandSched(illum_bins=illum_bins, loaded_band_groups=band_groups)

    def test_SimpleBandSched(self):
        bandsched = SimpleBandSched(illum_limit=40)
        brightmoon_result = ["g", "r", "i", "z", "y"]
        newmoon_result = ["u", "g", "r", "i", "z"]
        mjd = SURVEY_START_MJD
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == newmoon_result)
        conditions.moon_phase_sunset = 50
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == brightmoon_result)

    def test_DateSwapBandScheduler(self):
        # No specific arguments should be fine
        bandsched = DateSwapBandScheduler()
        # known time with known filters
        tt = Time("2025-08-05T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        self.assertEqual(
            bandsched(conditions),
            [
                "u",
                "r",
                "i",
                "z",
            ],
        )
        # full moon, past the current end of the band scheduler
        tt2 = Time("2025-11-05T20:00:00")
        conditions = Conditions(nside=8, mjd=tt2.mjd)
        conditions.moon_phase_sunset = 100
        self.assertEqual(bandsched(conditions), ["g", "r", "i", "z", "y"])
        # Set values - including override history
        date_swaps = {"2025-08-05": ["u", "g"], "2025-08-06": ["z", "y"]}
        end_date = Time("2025-08-07T12:00:00")
        backup_band_scheduler = SimpleBandSched(illum_limit=50)
        bandsched = DateSwapBandScheduler(
            swap_schedule=date_swaps, end_date=end_date, backup_band_scheduler=backup_band_scheduler
        )
        tt = Time("2025-08-05T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        self.assertEqual(bandsched(conditions), ["u", "g"])
        tt = Time("2025-08-06T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        self.assertEqual(bandsched(conditions), ["z", "y"])
        tt = Time("2025-08-07T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        conditions.moon_phase_sunset = 25
        self.assertEqual(bandsched(conditions), ["u", "g", "r", "i", "z"])
        # And just check that sorting doesn't matter in the input dicts
        date_swaps = {
            "2025-08-06": ["z", "y"],
            "2025-08-05": ["u", "g"],
        }
        end_date = Time("2025-08-07T12:00:00")
        backup_band_scheduler = SimpleBandSched(illum_limit=50)
        bandsched = DateSwapBandScheduler(
            swap_schedule=date_swaps, end_date=end_date, backup_band_scheduler=backup_band_scheduler
        )
        tt = Time("2025-08-05T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        self.assertEqual(bandsched(conditions), ["u", "g"])
        tt = Time("2025-08-06T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        self.assertEqual(bandsched(conditions), ["z", "y"])
        tt = Time("2025-08-07T20:00:00")
        conditions = Conditions(nside=8, mjd=tt.mjd)
        conditions.moon_phase_sunset = 25
        self.assertEqual(bandsched(conditions), ["u", "g", "r", "i", "z"])


if __name__ == "__main__":
    unittest.main()



import logging
import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.surveys import BaseSurvey


class NoObsSurvey(BaseSurvey):
    """Dummy class that always returns no valid reward"""

    def calc_reward_function(self, conditions):
        return -np.inf


class TestSimRunner(unittest.TestCase):

    def test_no_obs(self):
        """Check that sim ends even if we stop returning observations."""
        mjd_start = utils.SURVEY_START_MJD
        nside = 32
        survey_length = 1.5  # days

        scheduler = CoreScheduler([NoObsSurvey([], detailers=[])])
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        # Turn off noisy log warnings
        logging.disable(logging.CRITICAL)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )

        assert len(observations) == 0

        observatory, scheduler, observations, reward_df, obs_rewards_series = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None, record_rewards=True
        )

        assert len(observations) == 0


if __name__ == "__main__":
    unittest.main()



"""Second generation hourglass plotting classes."""

__all__ = ("riseset_times",)

import warnings

import astropy.coordinates
import astropy.time
import numpy as np
from astropy import units as u
from astropy.utils import iers


def riseset_times(  # pylint: disable=too-many-locals
    night_mjds,
    which_direction="down",
    which_night="nearest",
    alt=-14.0,
    location=astropy.coordinates.EarthLocation.of_site("Cerro Pachon"),
    body="sun",
    tolerance=1e-8,
    max_iter=10,
):
    """Find morning or evening twilight using Newton's iterative method

    This fails near the poles!

    Args:
        night_mjds : `numpy.array`
           MJDs of nights (integers)
        which_direction : `str`
           'evening' or 'morning'
        which_night : `str`
           'previous', 'nearest', or 'next'
        alt : `float`
            altitude of twilight, in degrees
        location : `astropy.coordinates.EarthLocation`
            location for site
        tolerance : `float`
            tolerance for twilight altitude, in degrees
        max_iter : `int`
            maximum iterations in Newton's method

    Returns:
        numpy array of mjds

    """

    # the astropy sidereal time routines often fail because they rely on
    # online updates to the IERS table, which is usually pointless
    # for this purpose. Use the formula from Meeus instead.
    def _compute_lst(times):
        try:
            lst = times.sidereal_time("apparent", longitude=location.lon)
        except (iers.IERSRangeError, ValueError):
            warnings.warn("Using (lower precision) Meeus formula for sidereal time.")
            mjd = times.mjd
            century = (mjd - 51544.5) / 36525
            gst = (
                280.46061837
                + 360.98564736629 * (mjd - 51544.5)
                + 0.000387933 * century * century
                - century * century * century / 38710000
            )
            lst = ((gst + location.lon.deg) % 360) * u.deg

        return lst

    event_direction = 1 if which_direction == "down" else -1

    night_wraps = {"previous": 0.0, "nearest": 180.0, "next": 360.0}
    night_wrap = night_wraps[which_night]

    mjds = night_mjds

    # Get close (to of order body motion per day)
    times = astropy.time.Time(mjds, scale="utc", format="mjd", location=location)
    lsts = _compute_lst(times)
    crds = astropy.coordinates.get_body(body, times, location=location)
    hour_angles = lsts - crds.ra
    event_hour_angles = event_direction * np.arccos(
        (np.sin(np.radians(alt)) - np.sin(crds.dec) * np.sin(location.lat))
        / (np.cos(crds.dec) * np.cos(location.lat))
    )
    event_hour_angles = astropy.coordinates.Angle(
        event_hour_angles, unit=u.radian  # pylint: disable=no-member
    )
    ha_diff = (event_hour_angles - hour_angles).wrap_at(night_wrap * u.deg)  # pylint: disable=no-member
    mjds = mjds + ha_diff.radian * (0.9972696 / (2 * np.pi))

    # Refine using Newton's method
    for iter_idx in range(max_iter):  # pylint: disable=unused-variable
        times = astropy.time.Time(mjds, scale="utc", format="mjd", location=location)
        crds = astropy.coordinates.get_body(body, times, location=location)
        current_alt = crds.transform_to(astropy.coordinates.AltAz(obstime=times, location=location)).alt
        finished = np.max(np.abs(current_alt.deg - alt)) < tolerance
        if finished:
            break

        current_sinalt = np.sin(current_alt.rad)
        target_sinalt = np.sin(np.radians(alt))

        ha = _compute_lst(times) - crds.ra  # pylint: disable=invalid-name
        # Derivative of the standard formula for sin(alt) in terms
        # of decl, latitude, and HA
        dsinalt_dlst = (-1 * np.cos(crds.dec) * np.cos(location.lat) * np.sin(ha)).value
        dsinalt_dmjd = dsinalt_dlst * (2 * np.pi / 0.9972696)
        mjds = mjds - (current_sinalt - target_sinalt) / dsinalt_dmjd

    if np.max(np.abs(mjds - night_mjds)) > 1:
        warnings.warn("On some nights, found twilight more than a day away from the night mjd")

    if not finished:
        unfinished = np.abs(current_alt.deg - alt) < tolerance
        if np.any(unfinished):
            try:
                first_unfinished_mjd = night_mjds[unfinished][0]
            except IndexError:
                first_unfinished_mjd = night_mjds[unfinished]
            warnings.warn(f"twilight_times did not converge, starting with {first_unfinished_mjd}")

    return mjds



__all__ = ("sim_runner",)

import copy
import gzip
import pickle
import sqlite3
import sys
import time
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.time import Time

from rubin_scheduler.scheduler.schedulers import SimpleBandSched
from rubin_scheduler.scheduler.utils import ObservationArray, SchemaConverter, run_info_table
from rubin_scheduler.utils import Site, _approx_altaz2pa, pseudo_parallactic_angle, rotation_converter


def sim_runner(
    observatory,
    scheduler,
    band_scheduler=None,
    sim_start_mjd=None,
    sim_duration=3.0,
    filename=None,
    delete_past=True,
    n_visit_limit=None,
    step_none=15.0,
    verbose=True,
    extra_info=None,
    event_table=None,
    record_rewards=False,
    start_result_size=int(2e5),
    append_result_size=int(2.5e6),
    anomalous_overhead_func=None,
    telescope="rubin",
    snapshot_dir="",
):
    """
    run a simulation

    Parameters
    ----------
    survey_length : float (3.)
        The length of the survey ot run (days)
    step_none : float (15)
        The amount of time to advance if the scheduler fails to
        return a target (minutes).
    extra_info : dict (None)
        If present, dict gets added onto the information from the
        observatory model.
    event_table : np.array (None)
        Any ToO events that were included in the simulation
    record_rewards : bool (False)
        Save computed rewards
    start_result_size : int
        Size of observations array to pre-allocate at the start of
        the run. Default 2e5.
    append_result_size : int
        Size of observations array to append if start_result_size is
        too small. Default 2.5e6.
    anomalous_overhead_func: `Callable` or None
        A function or callable object that takes the visit time and slew time
        (in seconds) as argument, and returns and additional offset (also
        in seconds) to be applied as addinional overhead between exposures.
        Defaults to None.
    telescope : `str`
        Name of telecope for camera rotation. Default "rubin".
    snapshot_dir : `str`
        Dir in which to safe scheduler + pickle snapshots with
        each scheduler call. An empty string if the snapshots should not be
        saved. Defaults to an empty string.
    """

    if extra_info is None:
        extra_info = {}

    t0 = time.time()

    if band_scheduler is None:
        band_scheduler = SimpleBandSched()

    if sim_start_mjd is None:
        mjd = observatory.mjd + 0
        sim_start_mjd = mjd + 0
    else:
        mjd = sim_start_mjd + 0
        observatory.mjd = mjd

    sim_end_mjd = sim_start_mjd + sim_duration
    observations = ObservationArray(n=start_result_size)
    mjd_track = mjd + 0
    step = 1.0 / 24.0
    step_none = step_none / 60.0 / 24.0  # to days
    mjd_run = sim_end_mjd - sim_start_mjd
    nskip = 0

    mjd_last_flush = -1

    last_obs_queue_fill_mjd_ns = None
    obs_rewards = {}
    reward_dfs = []

    rc = rotation_converter(telescope=telescope)

    # Make sure correct bands are mounted
    conditions = observatory.return_conditions()
    bands_needed = band_scheduler(conditions)
    observatory.observatory.mount_bands(bands_needed)

    counter = 0
    while mjd < sim_end_mjd:
        # Save pickle here
        if snapshot_dir is not None and len(snapshot_dir) > 0:
            snapshot_dt_isot = Time(mjd, format="mjd").isot
            # The assert is needed to make type hint checkers happy, because
            # they cannot tell whether Time.isot will return a sting or an
            # array of strings. But, mjd here is a single value, so
            # Time.isot will be as well.
            if isinstance(snapshot_dt_isot, np.ndarray):
                assert len(snapshot_dt_isot) == 1
                snapshot_dt_isot = snapshot_dt_isot.item()

            assert isinstance(snapshot_dt_isot, str)
            snapshot_dt = snapshot_dt_isot.replace(":", "")
            snapshot_fname = Path(snapshot_dir).joinpath(f"sched_snapshot_{snapshot_dt}.p.gz").as_posix()
            # Save before updating the conditions to mimic observatory.
            # There is a significant tradeoff in compression: even a little
            # compression makes the save both much smaller and much slower.
            # A simple test on a USDF devel node, running one night:
            # open method                     time     size
            # open                            3m18s    173G
            # gzip.open(compresslevel=1)      16m56s   19G
            # lzma.open(preset=1)             48m33s   8.7G
            # lzma.open(preset=9)             4h20m31s 2.7G
            # For equivilent times, lzma always seems better, but its lowest
            # compression level is slower than gzip's lowest compression
            # level.
            # I also explored bzip2, but for any given compression level,
            # it was always slower then either gzip or lzma (or both).
            with gzip.open(snapshot_fname, "wb", compresslevel=1) as pio:
                pickle.dump([scheduler, observatory.return_conditions()], file=pio)

        if not scheduler._check_queue_mjd_only(observatory.mjd):
            scheduler.update_conditions(observatory.return_conditions())

        desired_obs = scheduler.request_observation(mjd=observatory.mjd)
        if record_rewards:
            if last_obs_queue_fill_mjd_ns != scheduler.queue_fill_mjd_ns:
                reward_dfs.append(scheduler.queue_reward_df)
                last_obs_queue_fill_mjd_ns = scheduler.queue_fill_mjd_ns

        if desired_obs is None:
            # No observation. Just step into the future and try again.
            warnings.warn("No observation. Step into the future and trying again.")
            observatory.mjd = observatory.mjd + step_none
            scheduler.update_conditions(observatory.return_conditions())
            nskip += 1
            # Check we didn't skip so far we should end
            if observatory.mjd > sim_end_mjd:
                break
            else:
                continue
        completed_obs, new_night = observatory.observe(desired_obs)

        if completed_obs is not None:
            if anomalous_overhead_func is not None:
                observatory.mjd += (
                    anomalous_overhead_func(completed_obs["visittime"], completed_obs["slewtime"]) / 86400
                )

            scheduler.add_observation(completed_obs)
            observations[counter] = completed_obs[0]
            band_scheduler.add_observation(completed_obs)
            counter += 1
            if counter == observations.size:
                add_observations = ObservationArray(n=append_result_size)
                observations = np.concatenate([observations, add_observations])

            if record_rewards:
                obs_rewards[completed_obs[0]["mjd"]] = last_obs_queue_fill_mjd_ns
        else:
            # An observation failed to execute, usually it was outside
            # the altitude limits.
            if observatory.mjd == mjd_last_flush:
                raise RuntimeError(
                    "Scheduler has failed to provide a valid observation multiple times "
                    f" at time ({observatory.mjd} from survey {scheduler.survey_index}."
                )
            # if this is a first offence, might just be that targets set.
            # Flush queue and try to get some new targets.
            scheduler.flush_queue()
            mjd_last_flush = copy.deepcopy(observatory.mjd)

        if new_night:
            # find out what bands we want mounted
            conditions = observatory.return_conditions()
            bands_needed = band_scheduler(conditions)
            observatory.observatory.mount_bands(bands_needed)

        mjd = observatory.mjd + 0
        if verbose:
            if (mjd - mjd_track) > step:
                progress = np.max((mjd - sim_start_mjd) / mjd_run * 100)
                text = "\rprogress = %.2f%%" % progress
                sys.stdout.write(text)
                sys.stdout.flush()
                mjd_track = mjd + 0
        if n_visit_limit is not None:
            if counter == n_visit_limit:
                break
        # XXX--handy place to interupt and debug
        # if len(observations) > 25:
        #    import pdb ; pdb.set_trace()

    # trim off any observations that were pre-allocated but not used
    observations = observations[0:counter]

    # Compute alt,az,pa, rottelpos for observations
    # Only warn if it's a low-accuracy astropy conversion
    lsst = Site("LSST")

    if len(observations) > 0:
        # Using pseudo_parallactic_angle, see https://smtn-019.lsst.io/v/DM-44258/index.html
        pa, alt, az = pseudo_parallactic_angle(
            np.degrees(observations["RA"]),
            np.degrees(observations["dec"]),
            observations["mjd"],
            lon=lsst.longitude,
            lat=lsst.latitude,
            height=lsst.height,
        )
        observations["alt"] = np.radians(alt)
        observations["az"] = np.radians(az)
        observations["pseudo_pa"] = np.radians(pa)
        observations["rotTelPos"] = rc._rotskypos2rottelpos(
            observations["rotSkyPos"], observations["pseudo_pa"]
        )

        # Also include traditional parallactic angle
        pa = _approx_altaz2pa(observations["alt"], observations["az"], lsst.latitude_rad)
        observations["pa"] = pa

    runtime = time.time() - t0
    print("Skipped %i observations" % nskip)
    print("Flushed %i observations from queue for being stale" % scheduler.flushed)
    print("Completed %i observations" % len(observations))
    print("ran in %i min = %.1f hours" % (runtime / 60.0, runtime / 3600.0))
    if len(observations) > 0:
        if filename is not None:
            print("Writing results to ", filename)
        if filename is not None:
            info = run_info_table(observatory, extra_info=extra_info)
            converter = SchemaConverter()
            converter.obs2opsim(observations, filename=filename, info=info, delete_past=delete_past)
        if event_table is not None:
            df = pd.DataFrame(event_table)
            con = sqlite3.connect(filename)
            df.to_sql("events", con)
            con.close()

        if record_rewards:
            reward_df = pd.concat(reward_dfs)
            obs_rewards_series = pd.Series(obs_rewards)
            obs_rewards_series.index.name = "mjd"
            obs_rewards_series.name = "queue_fill_mjd_ns"

            if filename is not None:
                with sqlite3.connect(filename) as con:
                    reward_df.to_sql("rewards", con)
                    obs_rewards_series.to_sql("obs_rewards", con)
    else:
        # Make sure there is something to return if there are no
        # observations
        reward_df = None
        obs_rewards_series = None

    if record_rewards:
        result = observatory, scheduler, observations, reward_df, obs_rewards_series
    else:
        result = observatory, scheduler, observations

    return result



import unittest

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.basis_functions import (
    HealpixLimitedBasisFunctionMixin,
    SimpleArrayBasisFunction,
)
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.utils import DEFAULT_NSIDE


class SimpleArrayAtHpixBasisFunction(HealpixLimitedBasisFunctionMixin, SimpleArrayBasisFunction):
    pass


class TestHealpixLimitedBasisFunctionMixin(unittest.TestCase):
    def setUp(self):
        self.hpid = 2111
        random_seed = 6563
        self.nside = DEFAULT_NSIDE
        self.npix = hp.nside2npix(self.nside)
        self.rng = np.random.default_rng(seed=random_seed)
        self.all_values = self.rng.random(self.npix)
        self.value = self.all_values[self.hpid]
        self.conditions = Conditions(nside=self.nside, mjd=60200.2)

    def test_array_data(self):
        basis_function = SimpleArrayBasisFunction(self.all_values)
        test_values = basis_function(self.conditions)
        self.assertTrue(np.array_equal(test_values, self.all_values))

    def test_scalar_data(self):
        basis_function = SimpleArrayAtHpixBasisFunction(self.hpid, self.all_values)
        test_value = basis_function(self.conditions)
        self.assertEqual(test_value, self.value)

        feasible = basis_function(self.conditions)
        self.assertTrue(feasible)

    def test_infeasible_at_hpix(self):
        values_invalid_at_hpix = self.all_values.copy()
        values_invalid_at_hpix[self.hpid] = -np.inf
        basis_function = SimpleArrayAtHpixBasisFunction(self.hpid, values_invalid_at_hpix)
        feasible = basis_function.check_feasibility(self.conditions)
        self.assertFalse(feasible)


if __name__ == "__main__":
    unittest.main()



__all__ = ("UnscheduledDowntimeData", "UnscheduledDowntimeMoreY1Data")

import random

import numpy as np
from astropy.time import Time, TimeDelta

from . import Almanac


class UnscheduledDowntimeData:
    """Handle (and create) the unscheduled downtime information.

    Parameters
    ----------
    start_time : `astropy.time.Time`
        The time of the start of the simulation.
        The cloud database will be assumed to start on Jan 01 of the
        same year.
    seed : `int`, optional
        The random seed for creating the random nights of unscheduled
        downtime. Default 1516231120.
    start_of_night_offset : `float`, optional
        The fraction of a day to offset from MJD.0 to reach the
        defined start of a night ('noon' works).
        Default 0.16 (UTC midnight in Chile) - 0.5
        (minus half a day) = -0.34
    survey_length : `int`, optional
        The number of nights in the total survey. Default 3650*2.
    """

    MINOR_EVENT = {"P": 0.0137, "length": 1, "level": "minor event"}
    INTERMEDIATE_EVENT = {"P": 0.00548, "length": 3, "level": "intermediate event"}
    MAJOR_EVENT = {"P": 0.00137, "length": 7, "level": "major event"}
    CATASTROPHIC_EVENT = {"P": 0.000274, "length": 14, "level": "catastrophic event"}

    def __init__(
        self,
        start_time,
        seed=1516231120,
        start_of_night_offset=-0.34,
        survey_length=3650 * 2,
    ):
        self.seed = seed
        self.survey_length = survey_length
        self.start_of_night_offset = start_of_night_offset
        year_start = start_time.datetime.year
        self.night0 = Time("%d-01-01" % year_start, format="isot", scale="tai") + TimeDelta(
            start_of_night_offset, format="jd"
        )
        self.start_time = start_time

        # Scheduled downtime data is a np.ndarray of start
        # / end / activity for each scheduled downtime.
        self.downtime = None
        self.make_data()

    def __call__(self):
        """Return the array of unscheduled downtimes.

        Parameters
        ----------
        time : `astropy.time.Time`
            Time in the simulation for which to find the current downtime.

        Returns
        -------
        downtime : `np.ndarray`
            The array of all unscheduled downtimes, with keys for
            'start', 'end', 'activity',  corresponding to
            `astropy.time.Time`, `astropy.time.Time`, and `str`.
        """
        return self.downtime

    def _downtime_status(self, time):
        """Look behind the scenes at the downtime status/next values"""
        next_start = self.downtime["start"].searchsorted(time, side="right")
        next_end = self.downtime["end"].searchsorted(time, side="right")
        if next_start > next_end:
            current = self.downtime[next_end]
        else:
            current = None
        future = self.downtime[next_start:]
        return current, future

    def make_data(self):
        """Configure the set of unscheduled downtimes.

        This function creates the unscheduled downtimes based on a set
        of probabilities of the downtime type occurance.

        The random downtime is calculated using the following
        probabilities:

        minor event : remainder of night and next day = 5/365 days
        e.g. power supply failure
        intermediate : 3 nights = 2/365 days e.g. repair band
        mechanism, rotator, hexapod, or shutter
        major event : 7 nights = 1/2*365 days
        catastrophic event : 14 nights = 1/3650 days e.g. replace a raft
        """
        random.seed(self.seed)

        starts = []
        ends = []
        acts = []
        night = 0
        while night < self.survey_length:
            prob = random.random()
            if prob < self.CATASTROPHIC_EVENT["P"]:
                start_night = self.night0 + TimeDelta(night, format="jd")
                starts.append(start_night)
                end_night = start_night + TimeDelta(self.CATASTROPHIC_EVENT["length"], format="jd")
                ends.append(end_night)
                acts.append(self.CATASTROPHIC_EVENT["level"])
                night += self.CATASTROPHIC_EVENT["length"] + 1
                continue
            else:
                prob = random.random()
                if prob < self.MAJOR_EVENT["P"]:
                    start_night = self.night0 + TimeDelta(night, format="jd")
                    starts.append(start_night)
                    end_night = start_night + TimeDelta(self.MAJOR_EVENT["length"], format="jd")
                    ends.append(end_night)
                    acts.append(self.MAJOR_EVENT["level"])
                    night += self.MAJOR_EVENT["length"] + 1
                    continue
                else:
                    prob = random.random()
                    if prob < self.INTERMEDIATE_EVENT["P"]:
                        start_night = self.night0 + TimeDelta(night, format="jd")
                        starts.append(start_night)
                        end_night = start_night + TimeDelta(self.INTERMEDIATE_EVENT["length"], format="jd")
                        ends.append(end_night)
                        acts.append(self.INTERMEDIATE_EVENT["level"])
                        night += self.INTERMEDIATE_EVENT["length"] + 1
                        continue
                    else:
                        prob = random.random()
                        if prob < self.MINOR_EVENT["P"]:
                            start_night = self.night0 + TimeDelta(night, format="jd")
                            starts.append(start_night)
                            end_night = start_night + TimeDelta(self.MINOR_EVENT["length"], format="jd")
                            ends.append(end_night)
                            acts.append(self.MINOR_EVENT["level"])
                            night += self.MINOR_EVENT["length"] + 1
            night += 1
        self.downtime = np.array(
            list(zip(starts, ends, acts)),
            dtype=[("start", "O"), ("end", "O"), ("activity", "O")],
        )

    def total_downtime(self):
        """Return total downtime (in days).

        Returns
        -------
        total : `int`
            Total number of downtime days.
        """
        total = 0
        for td in self.downtime["end"] - self.downtime["start"]:
            total += td.jd
        return total


class UnscheduledDowntimeMoreY1Data(UnscheduledDowntimeData):
    def calc_sunrise_sets(self):
        """ """
        almanac = Almanac(mjd_start=self.start_time.mjd)
        year1_sunsets = np.where(
            (almanac.sunsets["night"] >= 0) & (almanac.sunsets["night"] < self.survey_length)
        )
        self.sunsets = almanac.sunsets[year1_sunsets]["sun_n12_setting"]
        self.sunrises = almanac.sunsets[year1_sunsets]["sun_n12_rising"]

    def make_data(self, end_of_start=380):
        """Configure the set of unscheduled downtimes.

        This function creates the unscheduled downtimes based on a set
        of probabilities of the downtime type occurance.

        The random downtime is calculated using the following
        probabilities:

        minor event : remainder of night and next day = 5/365 days
        e.g. power supply failure
        intermediate : 3 nights = 2/365 days e.g. repair band
        mechanism, rotator, hexapod, or shutter
        major event : 7 nights = 1/2*365 days
        catastrophic event : 14 nights = 1/3650 days e.g. replace a raft
        """

        self.calc_sunrise_sets()

        self.rng = np.random.default_rng(seed=self.seed)

        starts = []
        ends = []
        acts = []

        night_counted = np.zeros(len(self.sunsets))

        for night, (sunset, sunrise) in enumerate(zip(self.sunsets, self.sunrises)):
            prob = self.rng.random()
            hours_in_night = (sunrise - sunset) * 24.0
            if night_counted[night] == 1:
                continue

            if night < end_of_start:
                # Estimate a threshold probability of having some downtime
                # 50% at start, dropping until end_of_start, where it
                # should be .. 5%?
                nightly_threshold = 0.5 * (1 - night / (end_of_start + 45))
                if prob <= nightly_threshold:
                    # Generate an estimate of how long the downtime should be
                    prob_time = self.rng.gumbel(loc=1, scale=6, size=1)[0]
                    if prob_time >= hours_in_night:
                        prob_time = hours_in_night
                    if prob_time <= 1:
                        prob_time = 1.0
                    # And generate a starting time during the night for
                    # this event
                    tmax = hours_in_night - prob_time
                    if tmax <= 0:
                        starts.append(Time(sunset, format="mjd", scale="utc"))
                        ends.append(Time(sunrise, format="mjd", scale="utc"))
                        acts.append("Year1 Eng")
                    else:
                        offset = self.rng.uniform(low=sunset, high=sunset + tmax / 24.0)
                        starts.append(Time(offset, format="mjd", scale="utc"))
                        ends.append(Time(offset + prob_time / 24.0, format="mjd", scale="utc"))
                        acts.append("Year1 Eng")
                night_counted[night] = 1
                continue
        self.downtime = np.array(
            list(zip(starts, ends, acts)),
            dtype=[("start", "O"), ("end", "O"), ("activity", "O")],
        )

        regular_downtime = UnscheduledDowntimeData(
            start_time=self.start_time,
            seed=self.seed,
            start_of_night_offset=self.start_of_night_offset,
            survey_length=self.survey_length,
        )
        reg_dt = regular_downtime()
        self.downtime = np.concatenate([self.downtime, reg_dt])
        mjds = [row["start"].mjd for row in self.downtime]
        order = np.argsort(mjds)
        self.downtime = self.downtime[order]



import unittest
import warnings

import numpy as np

from rubin_scheduler.utils import Site


class SiteTest(unittest.TestCase):
    def setUp(self):
        # LSST default values taken from LSE-30
        self.height = 2650.0
        self.longitude = -70.74941
        self.latitude = -30.244628
        self.temperature = 11.5
        self.humidity = 0.4
        self.pressure = 750.0
        self.lapse_rate = 0.0065

    def test_lsst_values(self):
        """
        Test that LSST values are set correctly
        """
        site = Site(name="LSST")
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

    def test_auxtel_values(self):
        site = Site(name="Auxtel")
        self.assertEqual(site.name, "Auxtel")

    def test_no_defaults(self):
        """
        Test that, if name is not 'LSST', values are set to None
        """
        with warnings.catch_warnings(record=True) as ww:
            site = Site(name="bob")

        msg = str(ww[0].message)

        self.assertIn("longitude", msg)
        self.assertIn("latitude", msg)
        self.assertIn("temperature", msg)
        self.assertIn("pressure", msg)
        self.assertIn("height", msg)
        self.assertIn("lapse_rate", msg)
        self.assertIn("humidity", msg)

        self.assertEqual(site.name, "bob")
        self.assertIsNone(site.longitude)
        self.assertIsNone(site.longitude_rad)
        self.assertIsNone(site.latitude)
        self.assertIsNone(site.latitude_rad)
        self.assertIsNone(site.temperature)
        self.assertIsNone(site.temperature_kelvin)
        self.assertIsNone(site.pressure)
        self.assertIsNone(site.humidity)
        self.assertIsNone(site.lapse_rate)
        self.assertIsNone(site.height)

    def test_override_lss_tdefaults(self):
        """
        Test that, even if LSST is specified, we are capable of overriding
        defaults
        """
        site = Site(name="LSST", longitude=26.0)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, 26.0)
        self.assertEqual(site.longitude_rad, np.radians(26.0))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

        site = Site(name="LSST", latitude=88.0)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, 88.0)
        self.assertEqual(site.latitude_rad, np.radians(88.0))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

        site = Site(name="LSST", height=4.0)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, 4.0)

        site = Site(name="LSST", temperature=7.0)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, 7.0)
        self.assertEqual(site.temperature_kelvin, 280.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

        site = Site(name="LSST", pressure=14.0)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, 14.0)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

        site = Site(name="LSST", humidity=2.1)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, 2.1)
        self.assertEqual(site.lapse_rate, self.lapse_rate)
        self.assertEqual(site.height, self.height)

        site = Site(name="LSST", lapse_rate=3.2)
        self.assertEqual(site.name, "LSST")
        self.assertEqual(site.longitude, self.longitude)
        self.assertEqual(site.longitude_rad, np.radians(self.longitude))
        self.assertEqual(site.latitude, self.latitude)
        self.assertEqual(site.latitude_rad, np.radians(self.latitude))
        self.assertEqual(site.temperature, self.temperature)
        self.assertEqual(site.temperature_kelvin, self.temperature + 273.15)
        self.assertEqual(site.pressure, self.pressure)
        self.assertEqual(site.humidity, self.humidity)
        self.assertEqual(site.lapse_rate, 3.2)
        self.assertEqual(site.height, self.height)

    def test_partial_params(self):
        """
        test that unspecified parameters get set to None
        """
        with warnings.catch_warnings(record=True) as ww:
            site = Site(longitude=45.0, temperature=20.0)

        msg = str(ww[0].message)
        self.assertIn("latitude", msg)
        self.assertIn("height", msg)
        self.assertIn("pressure", msg)
        self.assertIn("lapse_rate", msg)
        self.assertIn("humidity", msg)
        self.assertNotIn("longitue", msg)
        self.assertNotIn("temperature", msg)

        self.assertIsNone(site.name)
        self.assertIsNone(site.latitude)
        self.assertIsNone(site.latitude_rad)
        self.assertIsNone(site.height)
        self.assertIsNone(site.pressure)
        self.assertIsNone(site.humidity)
        self.assertIsNone(site.lapse_rate)
        self.assertEqual(site.longitude, 45.0)
        self.assertEqual(site.longitude_rad, np.pi / 4.0)
        self.assertEqual(site.temperature, 20.0)
        self.assertEqual(site.temperature_kelvin, 293.15)

    def test_eq(self):
        """
        Test that we have correctly implemented __eq__ in Site
        """
        reference_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertEqual(reference_site, other_site)
        self.assertFalse(reference_site != other_site)
        self.assertTrue(reference_site == other_site)

        # just in case we ever change the class to convert
        # to radians only on demand, call for latitude/longitude
        # in radians and then check that the two instances are
        # still equal (since __eq__ just loops over the contents
        # of self.__dict__, this could fail if other_site has not
        # yet assigned a value to longitude/latitude_rad
        reference_site.latitude_rad
        self.assertEqual(reference_site, other_site)
        self.assertFalse(reference_site != other_site)
        self.assertTrue(reference_site == other_site)

        reference_site.longitude_rad
        self.assertEqual(reference_site, other_site)
        self.assertFalse(reference_site != other_site)
        self.assertTrue(reference_site == other_site)

        reference_site.temperature_kelvin
        self.assertEqual(reference_site, other_site)
        self.assertFalse(reference_site != other_site)
        self.assertTrue(reference_site == other_site)

        # now test that __ne__ works correctly
        other_site = Site(
            name="other",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.13,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.122,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.2,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.3,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.3,
            humidity=0.341,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.342,
            lapse_rate=0.008,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        other_site = Site(
            name="ref",
            longitude=112.12,
            latitude=-83.121,
            temperature=112.1,
            height=3124.2,
            pressure=891.2,
            humidity=0.341,
            lapse_rate=0.009,
        )

        self.assertNotEqual(reference_site, other_site)
        self.assertFalse(reference_site == other_site)
        self.assertTrue(reference_site != other_site)

        # test blank Sites
        ref_site = Site()
        other_site = Site()
        self.assertEqual(ref_site, other_site)
        self.assertTrue(ref_site == other_site)
        self.assertFalse(ref_site != other_site)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from rubin_scheduler.utils import SpecMap, default_spec_map


class SpecMapTest(unittest.TestCase):
    def verify_file(self, file_name, dir_name, test_spec_map=default_spec_map):
        """
        Verify that test_spec_map[file_name] results in
        os.path.join(dir_name, file_name+'.gz')
        """
        test_name = test_spec_map[file_name]
        control_name = os.path.join(dir_name, file_name + ".gz")
        msg = "%s should map to %s; it actually maps to %s" % (
            file_name,
            control_name,
            test_name,
        )
        self.assertEqual(test_name, control_name, msg=msg)

        add_space = file_name + " "
        self.assertNotEqual(add_space, file_name)
        test_name = test_spec_map[add_space]
        msg = "%s should map to %s; it actually maps to %s" % (
            add_space,
            control_name,
            test_name,
        )
        self.assertEqual(test_name, control_name, msg=msg)

        add_space = " " + file_name
        self.assertNotEqual(add_space, file_name)
        test_name = test_spec_map[add_space]
        msg = "%s should map to %s; it actually maps to %s" % (
            add_space,
            control_name,
            test_name,
        )
        self.assertEqual(test_name, control_name, msg=msg)

        add_gz = file_name + ".gz"
        self.assertNotEqual(add_gz, file_name)
        test_name = test_spec_map[add_gz]
        msg = "%s should map to %s; it actually maps to %s" % (
            add_gz,
            control_name,
            test_name,
        )
        self.assertEqual(test_name, control_name, msg=msg)

    def test_mlt(self):
        """
        Test that default_spec_map correctly locates MLT dwarf spectra
        """
        self.verify_file("lte004-3.5-0.0a+0.0.BT-Settl.spec", "starSED/mlt")

    def test_m_spec(self):
        """
        Test that default_spec_map correctly finds old MLT dwarf spectra
        that begin with 'm'
        """
        self.verify_file("m5.1Full.dat", "starSED/old_mlt")

    def test_l4_spec(self):
        """
        Test that default_spec_map correctly finds l4Full.dat
        """
        self.verify_file("l4Full.dat", "starSED/old_mlt")

    def test_l_spec(self):
        """
        Test that default_spec_map correctly find the L#_# spectra
        """
        self.verify_file("L2_0Full.dat", "starSED/old_mlt")

    def test_burrows_spec(self):
        """
        Test that default_spec_map correctly find the burrows spectra
        """
        self.verify_file("burrows+2006c91.21_T1400_g5.5_cf_0.3X", "starSED/old_mlt")

    def test_bergeron(self):
        """
        Test that default_spec_map correctly locates the bergeron spectra
        """
        self.verify_file("bergeron_4750_85.dat_4900", "starSED/wDs")

    def test_kurucz(self):
        """
        Test that default_spec_map correctly locates the kurucz spectra
        """
        self.verify_file("km30_5000.fits_g10_5040", "starSED/kurucz")
        self.verify_file("kp10_9000.fits_g40_9100", "starSED/kurucz")

    def test_galaxy(self):
        """
        Test that default_spec_map correctly locates the galaxy SEDs
        """
        self.verify_file("Const.79E06.002Z.spec", "galaxySED")
        self.verify_file("Inst.79E06.02Z.spec", "galaxySED")
        self.verify_file("Exp.40E08.02Z.spec", "galaxySED")
        self.verify_file("Burst.40E08.002Z.spec", "galaxySED")

    def test_dir_dict(self):
        """
        Test a user-defined SpecMap with a dir_dict
        """
        dir_dict_test_map = SpecMap(dir_dict={"(^lte)": "silly_sub_dir"})
        self.verify_file("lte_11111.txt", "silly_sub_dir", test_spec_map=dir_dict_test_map)
        self.verify_file("Const.79E06.002Z.spec", "galaxySED", test_spec_map=dir_dict_test_map)
        self.verify_file("Inst.79E06.02Z.spec", "galaxySED", test_spec_map=dir_dict_test_map)
        self.verify_file("Exp.40E08.02Z.spec", "galaxySED", test_spec_map=dir_dict_test_map)
        self.verify_file("Burst.40E08.002Z.spec", "galaxySED", test_spec_map=dir_dict_test_map)
        self.verify_file("km30_5000.fits_g10_5040", "starSED/kurucz", test_spec_map=dir_dict_test_map)
        self.verify_file("kp10_9000.fits_g40_9100", "starSED/kurucz", test_spec_map=dir_dict_test_map)
        self.verify_file(
            "burrows+2006c91.21_T1400_g5.5_cf_0.3X",
            "starSED/old_mlt",
            test_spec_map=dir_dict_test_map,
        )
        self.verify_file("L2_0Full.dat", "starSED/old_mlt", test_spec_map=dir_dict_test_map)
        self.verify_file("m5.1Full.dat", "starSED/old_mlt", test_spec_map=dir_dict_test_map)

    def test_file_dict(self):
        """
        Test a user-defined SpecMap with a file_dict
        """
        file_dict_test_map = SpecMap(file_dict={"abcd.txt": "file_dict_test_dir/abcd.txt.gz"})

        self.assertEqual(file_dict_test_map["abcd.txt"], "file_dict_test_dir/abcd.txt.gz")
        self.verify_file("lte_11111.txt", "starSED/mlt", test_spec_map=file_dict_test_map)
        self.verify_file("Const.79E06.002Z.spec", "galaxySED", test_spec_map=file_dict_test_map)
        self.verify_file("Inst.79E06.02Z.spec", "galaxySED", test_spec_map=file_dict_test_map)
        self.verify_file("Exp.40E08.02Z.spec", "galaxySED", test_spec_map=file_dict_test_map)
        self.verify_file("Burst.40E08.002Z.spec", "galaxySED", test_spec_map=file_dict_test_map)
        self.verify_file(
            "km30_5000.fits_g10_5040",
            "starSED/kurucz",
            test_spec_map=file_dict_test_map,
        )
        self.verify_file(
            "kp10_9000.fits_g40_9100",
            "starSED/kurucz",
            test_spec_map=file_dict_test_map,
        )
        self.verify_file(
            "burrows+2006c91.21_T1400_g5.5_cf_0.3X",
            "starSED/old_mlt",
            test_spec_map=file_dict_test_map,
        )
        self.verify_file("L2_0Full.dat", "starSED/old_mlt", test_spec_map=file_dict_test_map)
        self.verify_file("m5.1Full.dat", "starSED/old_mlt", test_spec_map=file_dict_test_map)

    def test_file_and_dir_dict(self):
        """
        Test a user-defined SpecMap with both a file_dict and a dir_dict
        """
        test_map = SpecMap(
            file_dict={"abcd.txt": "file_dir/abcd.txt.gz"},
            dir_dict={"(^burrows)": "dir_dir"},
        )

        self.assertEqual(test_map["abcd.txt"], "file_dir/abcd.txt.gz")
        self.verify_file("lte_11111.txt", "starSED/mlt", test_spec_map=test_map)
        self.verify_file("Const.79E06.002Z.spec", "galaxySED", test_spec_map=test_map)
        self.verify_file("Inst.79E06.02Z.spec", "galaxySED", test_spec_map=test_map)
        self.verify_file("Exp.40E08.02Z.spec", "galaxySED", test_spec_map=test_map)
        self.verify_file("Burst.40E08.002Z.spec", "galaxySED", test_spec_map=test_map)
        self.verify_file("km30_5000.fits_g10_5040", "starSED/kurucz", test_spec_map=test_map)
        self.verify_file("kp10_9000.fits_g40_9100", "starSED/kurucz", test_spec_map=test_map)
        self.verify_file("burrows+2006c91.21_T1400_g5.5_cf_0.3X", "dir_dir", test_spec_map=test_map)
        self.verify_file("L2_0Full.dat", "starSED/old_mlt", test_spec_map=test_map)
        self.verify_file("m5.1Full.dat", "starSED/old_mlt", test_spec_map=test_map)

    def test_contains(self):
        """
        Test that 'k in SpecMap' works as it should
        """

        test_map = SpecMap(
            file_dict={"abcd.txt": "file_dir/abcd.txt.gz"},
            dir_dict={"(^burrows)": "dir_dir"},
        )

        self.assertFalse("banana" in test_map)
        self.assertTrue("abcd.txt" in test_map)
        self.assertTrue("burrows_123.txt" in test_map)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import healpy as hp
import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.utils import (
    EuclidOverlapFootprint,
    PerFilterStep,
    SkyAreaGenerator,
    SkyAreaGeneratorGalplane,
    generate_all_sky,
    make_rolling_footprints,
)

datadir = os.path.join(get_data_dir(), "scheduler")


class TestSkyArea(unittest.TestCase):
    def setUp(self):
        self.nside = 32

    def test_perfilterstep(self):
        survey_length = 13
        pfs = PerFilterStep(
            loaded_dict={"u": np.array([0, 1, 2, 3, 10, 11])},
            survey_length=survey_length,
        )
        # check at day 1
        t1 = pfs(1, np.array([0]))

        # check at day 7
        t7 = pfs(7, np.array([0]))

        t_end = pfs(survey_length, np.array([0]))

        # u should grow fast at the start
        assert t1[0] > t1[2]

        # i and r should always grow the same
        assert t1[2] == t1[3]
        assert t7[2] == t7[3]

        # Everthing should be one at the end
        assert np.unique(t_end) == 1

    def test_make_rolling(self):
        """Check that we can make rolling footprints"""
        # Check that we can run with no kwargs
        footprints = make_rolling_footprints()

        assert footprints is not None

        # check that various kwarg combos work
        for nslice in [2, 3]:
            for uniform in [True, False]:
                for order_roll in [0, 1]:
                    for n_cycles in [3, 4]:
                        footprints = make_rolling_footprints(
                            nslice=nslice,
                            scale=0.8,
                            nside=32,
                            wfd_indx=None,
                            order_roll=order_roll,
                            n_cycles=n_cycles,
                            n_constant_start=2,
                            n_constant_end=6,
                            verbose=False,
                            uniform=uniform,
                        )

                        assert footprints is not None

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareagenerator(self):
        # Just test that it sets up and returns maps
        s = SkyAreaGenerator(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = ["", "LMC_SMC", "bulge", "dusty_plane", "lowdust", "nes", "scp", "virgo"]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareagenerator_nside(self):
        # Just check two other likely common nsides
        for nside in (16, 64):
            s = SkyAreaGenerator(nside=nside)
            footprints, labels = s.return_maps()
            lowdust = np.where(labels == "lowdust")
            self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareageneratorgalplane(self):
        # Just test that it sets up and returns maps
        s = SkyAreaGeneratorGalplane(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = ["", "LMC_SMC", "bulgy", "dusty_plane", "lowdust", "nes", "scp", "virgo"]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_euclidoverlapfootprint(self):
        # Just test that it sets up and returns maps
        s = EuclidOverlapFootprint(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = [
            "",
            "LMC_SMC",
            "bulgy",
            "dusty_plane",
            "lowdust",
            "nes",
            "scp",
            "virgo",
            "euclid_overlap",
        ]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    def test_generate_all_sky(self):
        # Test that the utility generate_all_sky returns appropriately
        nside = 32
        sky = generate_all_sky(nside=nside)
        expected_keys = ["map", "ra", "dec", "eclip_lat", "eclip_lon", "gal_lat", "gal_lon"]
        for k in expected_keys:
            self.assertTrue(k in sky.keys())
        self.assertEqual(sky["ra"].size, hp.nside2npix(nside))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from rubin_scheduler.scheduler.utils import smallest_signed_angle
from rubin_scheduler.utils import (
    SURVEY_START_MJD,
    Site,
    _approx_ra_dec2_alt_az,
    pseudo_parallactic_angle,
    rotation_converter,
)


class TestRotSkyConvert(unittest.TestCase):
    def test_rotation_converter(self):
        pa_vals = np.linspace(0, 360, 15)
        for tn in ["rubin", "auxtel", "comcam"]:
            rc = rotation_converter(telescope=tn)
            for pa in pa_vals:
                in_vals = np.linspace(-180, 180, 15)
                step1 = rc.rottelpos2rotskypos(in_vals, pa)
                step2 = rc.rotskypos2rottelpos(step1, pa)
                diff = np.abs(step2 - in_vals)
                pot_flip = np.where(diff > 1)
                diff[pot_flip] = diff[pot_flip] - 360
                assert np.allclose(diff, diff * 0)

                in_vals = np.linspace(0, 360, 15)
                step1 = rc.rotskypos2rottelpos(in_vals, pa)
                step2 = rc.rottelpos2rotskypos(step1, pa)
                diff = np.abs(step2 - in_vals)
                pot_flip = np.where(np.abs(diff) > 1)
                diff[pot_flip] = diff[pot_flip] - 360
                assert np.allclose(diff, diff * 0)
        # check that scalars work as well
        for tn in ["rubin", "auxtel", "comcam"]:
            rc = rotation_converter(telescope=tn)
            for pa in pa_vals:
                in_vals = np.linspace(-180, 180, 15)
                for iv in in_vals:
                    step1 = rc.rottelpos2rotskypos(iv, pa)
                    step2 = rc.rotskypos2rottelpos(step1, pa)
                    diff = abs(step2 - iv)
                    if diff > 1:
                        diff = diff - 360
                    assert np.allclose(diff, diff * 0)

                in_vals = np.linspace(0, 360, 15)
                for iv in in_vals:
                    step1 = rc.rotskypos2rottelpos(iv, pa)
                    step2 = rc.rottelpos2rotskypos(step1, pa)
                    diff = abs(step2 - iv)
                    if diff > 1:
                        diff = diff - 360
                    assert np.allclose(diff, diff * 0)

        with self.assertRaises(ValueError):
            rc = rotation_converter(telescope="not_a_telescope_name")

    def test_pseudo_pa(self):
        # Check that the pseudo parallactic angle is
        # somewhat close to the approx parallactic angle
        lsst = Site("LSST")
        rng = np.random.default_rng(seed=42)

        n = 100
        ra = rng.uniform(low=0, high=360, size=n)
        dec = rng.uniform(low=-90, high=90, size=n)
        mjd = np.arange(n) + SURVEY_START_MJD

        psudo_pa, salt, saz = pseudo_parallactic_angle(ra, dec, mjd, lon=lsst.longitude, height=lsst.height)

        falt, faz, fpa = _approx_ra_dec2_alt_az(
            np.radians(ra),
            np.radians(dec),
            np.radians(lsst.latitude),
            np.radians(lsst.longitude),
            mjd,
            return_pa=True,
        )

        diff = smallest_signed_angle(psudo_pa, np.degrees(fpa))
        diff[np.where(diff > 90)] -= 180
        diff[np.where(diff < -90)] += 180

        # Say they should be within 5 degrees
        assert np.max(np.abs(diff)) < 5.0


if __name__ == "__main__":
    unittest.main()



__all__ = ("SharedNumpyArray",)

from multiprocessing import shared_memory

import numpy as np


class SharedNumpyArray:
    """Class to hold a numpy array that can be shared with
    multiple processes without copying

    Origianlly from:
    https://e-dorigatti.github.io/python/2020/06/19/multiprocessing-large-objects.html

    Parameters
    ----------
    array : `np.array`
        Array to copy into shared memory.
    """

    def __init__(self, array):
        # Create the shared memory location of the same size of the array
        self._shared = shared_memory.SharedMemory(create=True, size=array.nbytes)
        # Save data type and shape, necessary to read the data correctly
        self._dtype, self._shape = array.dtype, array.shape
        # Create a new numpy array that uses the shared memory we created.
        res = np.ndarray(self._shape, dtype=self._dtype, buffer=self._shared.buf)
        # Copy data from the array to the shared memory. Numpy will
        # take care of copying everything in the correct format
        np.copyto(res, array)

    def read(self):
        """Read array without copy."""
        return np.ndarray(self._shape, self._dtype, buffer=self._shared.buf)

    def copy(self):
        """Copy array"""
        return np.copy(self.read())

    def unlink(self):
        """Unlink when done with data"""
        self._shared.close()
        self._shared.unlink()



__all__ = (
    "BaseFeature",
    "BaseSurveyFeature",
    "NObsCount",
    "LastObservation",
    "NObservations",
    "LastObserved",
    "NObsNight",
    "PairInNight",
    "NObservationsCurrentSeason",
    "LastNObsTimes",
    "NoteInNight",
    "LastObservationMjd",
    "LastObservedMatching",
)

import warnings
from abc import ABC, abstractmethod

import healpy as hp
import numpy as np
from scipy.stats import binned_statistic

from rubin_scheduler.scheduler import utils
from rubin_scheduler.scheduler.utils import IntRounded, ObservationArray
from rubin_scheduler.skybrightness_pre import dark_sky
from rubin_scheduler.utils import (
    DEFAULT_NSIDE,
    SURVEY_START_MJD,
    _angular_separation,
    _hpid2_ra_dec,
    calc_season,
)


def send_unused_deprecation_warning(name):
    message = (
        f"The feature {name} is not in use by the current "
        "baseline scheduler and may be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class BaseFeature(ABC):
    """The base class for features.
    This defines the standard API: a Feature should include
    a `self.feature` attribute, which could be a float, bool,
    or healpix size numpy array, or numpy masked array, and a
    __call__ method which returns `self.feature`.
    """

    def __init__(self, **kwargs):
        # self.feature should be a float, bool, or healpix size numpy
        # array, or numpy masked array
        self.feature = None

    def __call__(self):
        return self.feature


class BaseSurveyFeature(BaseFeature):
    """Track information relevant to the progress of a survey, using
    `self.feature` to hold this information.

    Features can track a single piece of information, or keep a map across
    the sky, or any other piece of information.

    Information in `self.feature` is updated via `add_observation` or
    `add_observation_array`.
    """

    @abstractmethod
    def add_observations_array(self, observations_array, observations_hpid):
        """Update self.feature based on information in `observations_array`
        and `observations_hpid`.

        This is a method to more rapidly restore the feature to its
        expected state, by using an array of all previous observations
        (`observations_array`) instead looping over the individual
        observations, as in `self.add_observation`. The observations array
        allows rapid calculation of acceptable observations, without
        considering factors such as which healpix is relevant.

        The `observations_hpid` is a reorganized version of the
        `observations_array`, where the array representing each observation
        is extended to include `hpid` and recorded multiple times (for each
        healpixel). The `observations_hpid` allows rapid calculation of
        feature values which depend on hpid.

        Links between `observations_array` and `observations_hpid` can
        be made using `ID`.
        """
        print(self)
        raise NotImplementedError

    @abstractmethod
    def add_observation(self, observation, indx=None, **kwargs):
        """Update self.feature based on information in `observation`.
        Observations should be ordered in monotonically increasing time.

        Parameters
        ----------
        observation : `np.array`, (1,N)
            Array of observation information, containing
            `mjd` for the time. See
            `rubin_scheduler.scheduler.utils.ObservationArray`.
        indx : `list`-like of [`int`]
            The healpixel indices that the observation overlaps.
            See `rubin_scheduler.utils.HpInLsstFov`.
        """
        raise NotImplementedError


class NoteInNight(BaseSurveyFeature):
    """Count appearances of any of `scheduler_notes` in
    observation `scheduler_note` in the
    current night; `note` must match one of `scheduler_notes`
    exactly.

    Useful for keeping track of how many times a survey or other subset
    of visits has executed in a given night.

    Parameters
    ----------
    notes : `list` [`str`], optional
        List of strings to match against observation `scheduler_note`
        values. The `scheduler_note` must match one of the items
        in notes exactly. Default of [None] will match any note.
    """

    def __init__(self, notes=[None]):
        self.feature = 0
        self.notes = notes
        self.current_night = -100

    def add_observations_array(self, observations_array, observations_hpid):
        # Identify the most recent night.
        if self.current_night != observations_array["night"][-1]:
            self.current_night = observations_array["night"][-1].copy()
            self.feature = 0
        # Identify observations within this night.
        indx = np.where(observations_array["night"] == observations_array["night"][-1])[0]
        # Count observations that match notes.
        for ind in indx:
            if (observations_array["scheduler_note"][ind] in self.notes) | (self.notes == [None]):
                self.feature += 1

    def add_observation(self, observation, indx=None):
        if self.current_night != observation["night"]:
            self.current_night = observation["night"].copy()
            self.feature = 0
        if (observation["scheduler_note"][0] in self.notes) | (self.notes == [None]):
            self.feature += 1


class NObsCount(BaseSurveyFeature):
    """Count the number of observations, whole sky (not per pixel).

    Because this feature will belong to a survey, it would count all
    observations that are counted for that survey.

    Parameters
    ----------
    scheduler_note : `str` or None
        Count observations that match `str` in their scheduler_note field.
        Note can be a substring of scheduler_note, and will still match.
    bandname : `str` or None
        Optionally also (or independently) specify a band to match.
    note : `str` or None
        Backwards compatible shim for scheduler_note.
    per_night : `bool`
        If True, only count within a night. Default False.
    """

    def __init__(self, scheduler_note=None, bandname=None, note=None, per_night=False):
        self.feature = 0
        if scheduler_note is None:
            self.scheduler_note = note
        else:
            self.scheduler_note = scheduler_note
        # In case scheduler_note got set with empty string, replace with None
        if self.scheduler_note == "":
            self.scheduler_note = None
        self.bandname = bandname
        # Will not add filtername backup here; users should not reach
        # down as far as the features in configurations.
        self.per_night = per_night
        self.night = -1

    def add_observations_array(self, observations_array, observations_hpid):
        if self.per_night:
            self.night = observations_array["night"].max() + 0
            observations_array = observations_array[
                np.where(observations_array["night"] == observations_array["night"].max())
            ]
        if self.scheduler_note is None and self.bandname is None:
            self.feature += observations_array.size
        elif self.scheduler_note is None and self.bandname is not None:
            in_filt = np.where(observations_array["band"] == self.bandname)
            self.feature += np.size(in_filt)
        elif self.scheduler_note is not None and self.bandname is None:
            count = [self.scheduler_note in note for note in observations_array["scheduler_note"]]
            self.feature += np.sum(count)
        else:
            # note and bandname are defined
            in_filt = np.where(observations_array["band"] == self.bandname)
            count = [self.scheduler_note in note for note in observations_array["scheduler_note"][in_filt]]
            self.feature += np.sum(count)

    def add_observation(self, observation, indx=None):
        if self.per_night:
            if observation["night"] != self.night:
                self.feature = 0
                self.night = observation["night"] + 0
        # Track all observations
        if self.scheduler_note is None and self.bandname is None:
            self.feature += 1
        elif self.scheduler_note is None and self.bandname is not None:
            if observation["band"][0] in self.bandname:
                self.feature += 1
        elif self.scheduler_note is not None and self.bandname is None:
            if self.scheduler_note in observation["scheduler_note"][0]:
                self.feature += 1
        else:
            if (observation["band"][0] in self.bandname) and self.scheduler_note in observation[
                "scheduler_note"
            ][0]:
                self.feature += 1


class LastObservation(BaseSurveyFeature):
    """Track the (entire) last observation.

    All visits, no healpix dependency.

    Parameters
    ----------
    scheduler_note : `str` or None, optional
        The scheduler_note to match.
        Scheduler_note values which match this OR which contain this value
        as a subset of their string will match.
    bandname : `str` or None, optional
        The required band to match.
    survey_name : `str` or None, optional
        Backwards compatible shim for scheduler_note. Deprecated.
    """

    def __init__(self, scheduler_note=None, bandname=None, survey_name=None):
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.bandname = bandname
        # Start out with an empty observation
        self.feature = utils.ObservationArray()

    def add_observations_array(self, observations_array, observations_hpid):
        valid_indx = np.ones(observations_array.size, dtype=bool)
        if self.bandname is not None:
            valid_indx[np.where(observations_array["band"] != self.bandname)[0]] = False
        if self.scheduler_note is not None:
            tmp = [self.scheduler_note in name for name in observations_array["scheduler_note"]]
            valid_indx = valid_indx * np.array(tmp)
        if valid_indx.sum() > 0:
            self.feature = observations_array[valid_indx][-1]

    def add_observation(self, observation, indx=None):
        if (self.scheduler_note is None) or (self.scheduler_note in observation["scheduler_note"][0]):
            if (self.bandname is None) or (self.bandname == observation["band"][0]):
                self.feature = observation


class NObservations(BaseSurveyFeature):
    """
    Track the number of observations that have been made at each healpix.

    Parameters
    ----------
    bandname : `str` or `list` [`str`] or None
        String or list that has all the bands that can count.
        Default None counts all bands.
    nside : `int`
        The nside of the healpixel map to use.
        Default None uses scheduler default.
    scheduler_note : `str` or None, optional
        The scheduler_note to match.
        Scheduler_note values which match this OR which contain this value
        as a subset of their string will match.
    survey_name : `str` or None
        The scheduler_note value to match.
        Deprecated in favor of scheduler_note, but provided for backward
        compatibility. Will be removed in the future.
    seeing_limit : `float`
        The limit to apply on accepting image quality.
    """

    def __init__(
        self, bandname=None, nside=DEFAULT_NSIDE, scheduler_note=None, survey_name=None, seeing_limit=None
    ):
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.bandname = bandname
        self.seeing_limit = seeing_limit
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.seeing_limit is not None:
            valid_indx[np.where(IntRounded(observations_hpid["FWHMeff"]) > IntRounded(self.seeing_limit))] = (
                False
            )
        if self.bandname is not None:
            valid_indx[np.where(observations_hpid["band"] != self.bandname)[0]] = False
        if self.scheduler_note is not None:
            tmp = [name in self.scheduler_note for name in observations_hpid["scheduler_note"]]
            valid_indx = valid_indx * np.array(tmp)
        data = observations_hpid[valid_indx]
        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(
                data["hpid"], np.ones(data.size), statistic=np.sum, bins=self.bins
            )
            self.feature += result

    def add_observation(self, observation, indx=None):
        seeing_check = True
        if self.seeing_limit is not None:
            seeing_check = IntRounded(observation["FWHMeff"]) <= IntRounded(self.seeing_limit)

        if self.bandname is None or observation["band"][0] in self.bandname:
            if seeing_check:
                if self.scheduler_note is None or observation["scheduler_note"][0] in self.scheduler_note:
                    self.feature[indx] += 1


class LargestN:
    def __init__(self, n):
        # This is used within other features or basis functions,
        # but is not a feature itself
        self.n = n

    def __call__(self, in_arr):
        if np.size(in_arr) < self.n:
            return 0
        result = in_arr[-self.n]
        return result


class LastNObsTimes(BaseSurveyFeature):
    """Record the last three observations for each healpixel.

    Parameters
    ----------
    bandname : `str` or None
        Match visits in this bandname.
    n_obs : `int`
        The number of prior observations to track.
    nside : `int`
        The nside of the healpix map.
    """

    def __init__(self, bandname=None, n_obs=3, nside=DEFAULT_NSIDE):
        self.bandname = bandname
        self.n_obs = n_obs
        min_dim = max(1, self.n_obs)
        self.feature = np.zeros((min_dim, hp.nside2npix(nside)), dtype=float)
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.bandname is not None:
            valid_indx[np.where(observations_hpid["band"] != self.bandname)[0]] = False
        data = observations_hpid[valid_indx]

        if np.size(data) > 0:
            # If self.n_obs is very large, this is a bad idea, but
            # our use is intended to be around n_obs 3.
            for i in range(1, self.n_obs + 1):
                func = LargestN(i)
                result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=func, bins=self.bins)
                self.feature[-i, :] = result

    def add_observation(self, observation, indx=None):
        if self.bandname is None or observation["band"][0] in self.bandname:
            self.feature[0:-1, indx] = self.feature[1:, indx]
            self.feature[-1, indx] = observation["mjd"]


class NObservationsCurrentSeason(BaseSurveyFeature):
    """Count the observations at each healpix, taken in the most
    recent season, that meet band, seeing and m5 criteria.

    Useful for ensuring that "good quality" observations are acquired
    in each season at each point in the survey footprint.

    Parameters
    ----------
    bandname : `str`, optional
        If None (default) count observations in any band.
        Otherwise, only count observations in the specified band.
    nside : `int`, optional
        If None (default), use default nside for scheduler.
        Otherwise, set nside for the map to nside.
    seeing_fwhm_max : `float`, optional
        If None (default), count observations up to any seeing value.
        Otherwise, only count observations with better seeing (`FWHMeff`)
        than `seeing_fwhm_max`. In arcseconds.
    m5_penalty_max : `float`, optional
        If None (default), count observations with any m5 value.
        Otherwise, only count observations within this value of the
        dark sky map at this pixel.
        Only relevant if bandname is not None.
    mjd_start : `float`, optional
        If None, uses default survey_start_mjd for the start of the survey.
        This defines the starting year for counting seasons, so should
        be the start of the survey.
    """

    def __init__(
        self,
        bandname=None,
        nside=DEFAULT_NSIDE,
        seeing_fwhm_max=None,
        m5_penalty_max=None,
        mjd_start=SURVEY_START_MJD,
    ):
        self.nside = nside
        self.seeing_fwhm_max = seeing_fwhm_max
        self.bandname = bandname
        self.m5_penalty_max = m5_penalty_max
        # If bandname not set, then we can't set m5_penalty_max.
        if self.bandname is None and self.m5_penalty_max is not None:
            warnings.warn("To use m5_penalty_max, bandname must be set. Disregarding m5_penalty_max.")
            self.m5_penalty_max = None

        # Get the dark sky map
        if self.bandname is not None:
            self.dark_map = dark_sky(nside)[bandname]
        self.ones = np.ones(hp.nside2npix(self.nside))
        self.mjd_start = mjd_start

        # Set up feature values - this includes the count in a given season
        # Find the healpixels for each point on the sky
        self.ra, self.dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
        self.ra_deg = np.degrees(self.ra)

        self.mjd = mjd_start
        # Set up season_map. This should be the same as conditions.season_map
        # but the important thing is that mjd_start matches.
        self.season_map = calc_season(self.ra_deg, [self.mjd], self.mjd_start).flatten()
        self.season = np.floor(self.season_map)

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        # Set bins for add_observations_array
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def season_update(self, observation=None, conditions=None):
        """Update the season_map to the current time.

        This assumes time increases monotonically in the conditions or
        observations objects passed as arguments.
        Using the 'mjd' from the conditions, where parts of the sky
        have changed season (increasing in the `self.season_map` + 1)
        the `self.feature` is cleared, in order to restart counting
        observations in the new season.

        Parameters
        ----------
        observation : `np.array`, (1,N)
            Array of observation information, containing
            `mjd` for the time. See
            `rubin_scheduler.scheduler.utils.ObservationArray`.
        conditions : `rubin_scheduler.scheduler.Conditions`, optional
            A conditions object, containing `mjd`.

        Notes
        -----
        One of observations or conditions must be passed, but either
        are options so this can be used with add_observation.
        Most of the time, the updates will come from `conditions`.
        """
        mjd_now = None
        mjd_from = None
        if observation is not None:
            mjd_now = observation["mjd"]
            mjd_from = "observation"
        # Prefer to use Conditions for the time.
        if conditions is not None:
            mjd_now = conditions.mjd
            mjd_from = "conditions"
        if mjd_now is None:
            warnings.warn(
                "Expected either a conditions or observations object. Not updating season_map.",
                UserWarning,
            )
            return
        if isinstance(mjd_now, np.ndarray) or isinstance(mjd_now, list):
            mjd_now = mjd_now[0]
        # Check that time doesn't go backwards.
        # But only bother to warn if it's coming from conditions.
        # Observations may look like they go backwards.
        if mjd_now < self.mjd:
            if mjd_from == "conditions":
                warnings.warn(
                    f"Time must flow forwards to track the "
                    f"feature in {self.__class__.__name__}."
                    f"Not updating season_map.",
                    UserWarning,
                )
            # But just return without update in either case.
            return
        # Calculate updated season values.
        updated_season = np.floor(self.season_map + (mjd_now - self.mjd_start) / 365.25)
        # Clear feature where season increased by 1 (note 'floor' above).
        self.feature[np.where(updated_season > self.season)] = 0
        # Update to new time and season.
        self.mjd = mjd_now
        self.season = updated_season

    def add_observations_array(self, observations_array, observations_hpid):
        # Update self.season to the most recent observation time
        self.season_update(observation=observations_array[-1])

        # Start assuming 'check' is all True, for all observations.
        check = np.ones(observations_array.size, dtype=bool)

        # Rule out the observations with seeing fwhmeff > limit
        if self.seeing_fwhm_max is not None:
            check[np.where(observations_array["FWHMeff"] > self.seeing_fwhm_max)] = False

        # Rule out observations which are not in the desired band
        if self.bandname is not None:
            check[np.where(observations_array["band"] != self.bandname)] = False

        # Convert the "check" array on observations_array into a
        # "check" array on the hpids so we can evaluate healpix-level items
        check_hpid_indxs = np.isin(observations_hpid["ID"], observations_array["ID"][check])
        # Set up a new valid/not valid flag, now on observations_hpid
        check_hp = np.zeros(len(observations_hpid), dtype=bool)
        # Set all observations which passed simpler tests above to True
        check_hp[check_hpid_indxs] = True

        hpids = observations_hpid["hpid"]

        # This could be done once per observation, but to make life
        # easier for index matching, let's just calculate season
        # based on the hpid array.
        # We only want to count observations that would fall within the
        # current season, so calculate the season for each obs.
        seasons = np.floor(
            calc_season(
                np.degrees(observations_hpid["RA"]),
                observations_hpid["mjd"],
                self.mjd_start,
                calc_diagonal=True,
            )
        )
        season_change = self.season[hpids] - seasons
        check_hp = np.where(season_change != 0, False, check_hp)
        # And check for m5_penalty_map
        if self.m5_penalty_max is not None:
            penalty = self.dark_map[hpids] - observations_hpid["fivesigmadepth"]
            check_hp = np.where(penalty > self.m5_penalty_max, False, check_hp)

        # Bin up the observations per hpid and add them to the feature.
        if np.sum(observations_hpid["hpid"][check_hp]) > 0:
            result, _be, _bn = binned_statistic(
                observations_hpid["hpid"][check_hp],
                observations_hpid["hpid"][check_hp],
                bins=self.bins,
                statistic=np.size,
            )
        else:
            result = 0
        # Add the resulting observations to feature.
        self.feature += result

    def add_observation(self, observation, indx):
        # Update the season map to the current time.
        self.season_update(observation=observation)
        # Check if *this* observation is in the current season
        # (This means we could add observations from the past, but
        # would count it against the current season).
        season_obs = np.floor(self.season_map[indx] + (observation["mjd"] - self.mjd_start) / 365.25)
        this_season_indx = np.array(indx)[np.where(season_obs == self.season[indx])]

        # Check if seeing is good enough.
        if self.seeing_fwhm_max is not None:
            check = observation["FWHMeff"] <= self.seeing_fwhm_max
        else:
            check = True

        # Check if observation is in the right band
        if self.bandname is not None and check:
            if observation["band"][0] != self.bandname:
                check = False
            else:
                # Check if observation in correct band and deep enough.
                if self.m5_penalty_max is not None:
                    penalty = self.dark_map[this_season_indx] - observation["fivesigmadepth"]
                    this_season_indx = this_season_indx[np.where(penalty <= self.m5_penalty_max)]

        if check:
            self.feature[this_season_indx] += 1


class LastObservedMatching(BaseSurveyFeature):
    """Record the last observation that matches a criteria"""

    def __init__(self, ang_distance_match=None, science_program=None, scheduler_note=None, ra=None, dec=None):
        # set starting feature to something
        self.feature = ObservationArray(n=1)
        self.feature.mjd = -100000
        self.science_program = science_program
        self.scheduler_note = scheduler_note
        if (ra is not None) | (dec is not None) | (ang_distance_match is not None):
            self.ra = np.radians(ra)
            self.ang_distance_match = np.radians(ang_distance_match)
            self.dec = np.radians(dec)
        else:
            self.ang_distance_match = None

    def add_observations_array(self, observations_array, observations_hpid):
        if self.science_program is not None:
            good = np.where(observations_array["science_program"] == self.science_program)[0]
            observations_array = observations_array[good]

        if self.scheduler_note is not None:
            good = np.where(observations_array["scheduler_note"] == self.scheduler_note)[0]
            observations_array = observations_array[good]

        if self.ang_distance_match is not None:
            dist = _angular_separation(observations_array["RA"], observations_array["dec"], self.ra, self.dec)
            good = np.where(dist <= self.ang_distance_match)
            observations_array = observations_array[good]

        if np.size(observations_array) > 0:
            self.feature = ObservationArray(n=1)
            self.feature[0] = observations_array[-1]

    def add_observation(self, observation, indx=None):
        if self.science_program is not None:
            if not self.science_program == observation["science_program"]:
                return

        if self.scheduler_note is not None:
            if not self.scheduler_note == observation["scheduler_note"]:
                return

        if self.ang_distance_match is not None:
            dist = _angular_separation(observation["RA"], observation["dec"], self.ra, self.dec)
            if dist >= self.ang_distance_match:
                return

        # If we make it this far, record the observation as matching
        self.feature = observation.copy()


class LastObserved(BaseSurveyFeature):
    """
    Track the MJD when a pixel was last observed.
    Assumes observations are added in chronological order.

    Calculated per healpix.

    Parameters
    ----------
    bandname : `str` or None
        Track visits in a particular band or any band (None).
    nside : `int` or None
        Nside for the healpix map, default of None uses the scheduler default.
    fill : `float`
        Fill value to use where no observations have been found.
    """

    def __init__(self, bandname="r", nside=DEFAULT_NSIDE, fill=np.nan):
        self.bandname = bandname
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float) + fill
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.bandname is not None:
            valid_indx[np.where(observations_hpid["band"] != self.bandname)[0]] = False
        data = observations_hpid[valid_indx]

        result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=np.max, bins=self.bins)
        good = np.where(result > 0)
        self.feature[good] = result[good]

    def add_observation(self, observation, indx=None):
        if self.bandname is None:
            self.feature[indx] = observation["mjd"]
        elif observation["band"][0] in self.bandname:
            self.feature[indx] = observation["mjd"]


class LastObservationMjd(BaseSurveyFeature):
    """Track the MJD of the last observation.

    All visits, no healpix dependency.

    Parameters
    ----------
    scheduler_note : `str` or None, optional
        The scheduler_note to match.
        Scheduler_note values which match this OR which contain this value
        as a subset of their string will match.
    bandname : `str` or None, optional
        The required band to match.

    Notes
    -----
    Very similar to LastObservation, but only tracks the MJD of the
    last matching visit.
    """

    def __init__(self, scheduler_note=None, bandname=None, note=None):
        self.scheduler_note = scheduler_note
        if self.scheduler_note is None and note is not None:
            self.scheduler_note = note
        self.bandname = bandname
        self.feature = None

    def add_observations_array(self, observations_array, observations_hpid):
        valid_indx = np.ones(observations_array.size, dtype=bool)
        if self.bandname is not None:
            valid_indx[np.where(observations_array["band"] != self.bandname)[0]] = False
        if self.scheduler_note is not None:
            tmp = [self.scheduler_note in name for name in observations_array["scheduler_note"]]
            valid_indx = valid_indx * np.array(tmp)
        if valid_indx.sum() > 0:
            self.feature = observations_array[valid_indx][-1]["mjd"]

    def add_observation(self, observation, indx=None):
        if (self.scheduler_note is None) or (self.scheduler_note in observation["scheduler_note"][0]):
            if (self.bandname is None) or (self.bandname == observation["band"][0]):
                self.feature = observation["mjd"]


class NObsNight(BaseSurveyFeature):
    """
    Track how many times a healpixel has been observed in a night.

    Parameters
    ----------
    bandname : `str` or None
        Filter to track. None tracks observations in any band.
    nside : `int` or None
        Scale of the healpix map. Default of None uses the scheduler
        default nside.
    """

    def __init__(self, bandname="r", nside=DEFAULT_NSIDE):
        self.bandname = bandname
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.night = None
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        latest_night = observations_array["night"].max()
        if latest_night != self.night:
            self.feature *= 0
            self.night = latest_night
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.bandname is not None:
            valid_indx[np.where(observations_hpid["band"] != self.bandname)[0]] = False
        if self.night is not None:
            valid_indx[np.where(observations_hpid["night"] != self.night)] = False

        data = observations_hpid[valid_indx]
        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(
                data["hpid"], np.ones(data.size), statistic=np.sum, bins=self.bins
            )
            self.feature += result

    def add_observation(self, observation, indx=None):
        if observation["night"] != self.night:
            self.feature *= 0
            self.night = observation["night"]
        if (self.bandname == "") | (self.bandname is None):
            self.feature[indx] += 1
        elif observation["band"][0] in self.bandname:
            self.feature[indx] += 1


class PairInNight(BaseSurveyFeature):
    """
    Track how many pairs have been observed within a night at a given healpix.

    Parameters
    ----------
    gap_min : `float` (25.)
        The minimum time gap to consider a successful pair in minutes
    gap_max : `float` (45.)
        The maximum time gap to consider a successful pair (minutes)
    """

    def __init__(self, bandname="r", nside=DEFAULT_NSIDE, gap_min=25.0, gap_max=45.0):
        self.bandname = bandname
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.indx = np.arange(self.feature.size)
        self.last_observed = LastObserved(bandname=bandname)
        self.gap_min = IntRounded(gap_min / (24.0 * 60))  # Days
        self.gap_max = IntRounded(gap_max / (24.0 * 60))  # Days
        self.night = 0
        # Need to keep a full record of times and healpixels observed in
        # a night.
        self.mjd_log = []
        self.hpid_log = []

    def add_observations_array(self, observations_array, observations_hpid):
        # ok, let's just find the largest night and toss all those in one
        # at a time
        # THIS IGNORES FILTER??
        most_recent_night = np.where(observations_hpid["night"] == np.max(observations_hpid["night"]))[0]
        obs_hpid = observations_hpid[most_recent_night]
        uid = np.unique(obs_hpid["ID"])
        for ind_id in uid:
            # maybe a faster searchsorted way to do this, but it'll work
            # for now
            good = np.where(obs_hpid["ID"] == ind_id)[0]
            self.add_observation(observations_hpid[good][0], observations_hpid[good]["hpid"])

    def add_observation(self, observation, indx=None):
        if self.bandname is None:
            infilt = True
        else:
            infilt = observation["band"][0] in self.bandname
        if infilt:
            if indx is None:
                indx = self.indx
            # Clear values if on a new night
            if self.night != observation["night"]:
                self.feature *= 0.0
                self.night = observation["night"]
                self.mjd_log = []
                self.hpid_log = []

            # Look for the mjds that could possibly pair with observation
            mjd_diff = IntRounded(observation["mjd"] - np.array(self.mjd_log))
            # normally would use np.searchsorted, but need to use IntRounded
            # to be sure we are cross-platform repeatable.
            in_range_indx = np.where((mjd_diff > self.gap_min) & (mjd_diff < self.gap_max))[0]
            # Now check if any of the healpixels taken in the time gap
            # match the healpixels of the observation.
            if in_range_indx.size > 0:
                left = in_range_indx.min()
                right = in_range_indx.max() + 1
                matches = np.isin(indx, self.hpid_log[left:right])
                self.feature[np.array(indx)[matches]] += 1

            # record the mjds and healpixels that were observed
            self.mjd_log.extend([np.max(observation["mjd"])] * np.size(indx))
            self.hpid_log.extend(list(indx))
